"""Tests for the web utility functions."""
from functools import partial
import http
import logging
import re
from typing import Any, Awaitable, Callable, Final, Optional, TypedDict, cast
from unittest.mock import AsyncMock, Mock, PropertyMock, call

from _pytest.logging import LogCaptureFixture
import httpx
import pytest
from pytest import fixture
from pytest_httpx import HTTPXMock
from pytest_mock import MockerFixture
import requests
from requests import Response
import responses
from typing_extensions import NotRequired

from bitfount import web_utils
from bitfount.web_utils import _auto_retry_request
from tests.utils import PytestRequest
from tests.utils.helper import get_debug_logs, unit_test

_HTTP_STATUS_CODES: Final = tuple(i.value for i in http.HTTPStatus)
_HTTP_CALL_TYPES: Final = (
    "request-head",
    "request-get",
    "request-post",
    "request-put",
    "request-patch",
    "request-delete",
    "head",
    "get",
    "post",
    "put",
    "patch",
    "delete",
)
_DEFAULT_BACKOFF_TIMES: Final = tuple(
    web_utils._DEFAULT_BACKOFF_FACTOR * (2**i)
    for i in range(web_utils._DEFAULT_MAX_RETRIES)
)


@fixture
def http_func_details(request: PytestRequest) -> str:
    """Simple fixture to extract the param from the request."""
    return str(request.param)


@fixture
def http_method(http_func_details: str) -> str:
    """The HTTP method from the details."""
    return http_func_details.split("-")[-1].upper()


@fixture
def test_url() -> str:
    """Test URL to use."""
    return "http://example.com"


@unit_test
class TestSyncWebUtils:
    """Tests for synchronous web interactions."""

    @fixture
    def http_func(
        self, http_func_details: str, http_method: str
    ) -> Callable[..., Response]:
        """The actual HTTP function to use.

        This will either be the syntactic sugar version (i.e. `.get(...)`) or the direct
        version (i.e. `.request(method="GET", ...)` depending on the form of
        http_func_details.
        """
        if http_func_details.startswith("request"):
            return partial(web_utils.request, method=http_method)
        else:
            return cast(
                Callable[..., Response], getattr(web_utils, http_method.lower())
            )

    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=_HTTP_CALL_TYPES,
        indirect=True,
    )
    @pytest.mark.parametrize("status_code", _HTTP_STATUS_CODES)
    @responses.activate
    def test_retry_handles_http_status_codes(
        self,
        http_func: Callable[..., Response],
        http_method: str,
        remove_web_retry_backoff_sleep: Optional[Mock],
        status_code: int,
        test_url: str,
    ) -> None:
        """Test that auto retry occurs for expected HTTP status codes.

        Checks for all status codes and HTTP method combinations.
        """
        # Check retry sleep is patched
        assert remove_web_retry_backoff_sleep is not None

        responses.add(
            url=test_url,
            status=status_code,
            method=http_method,
        )

        resp = http_func(url=test_url)

        assert resp.status_code == status_code
        assert resp.request.method == http_method

        # Construct expected values depending on if retryable
        if status_code in web_utils._RETRY_STATUS_CODES:
            expected_req_count = 1 + web_utils._DEFAULT_MAX_RETRIES
            expected_sleep_calls = [call(i) for i in _DEFAULT_BACKOFF_TIMES]
        else:
            expected_req_count = 1
            expected_sleep_calls = []

        # Check if retries occurred as expected
        responses.assert_call_count(test_url, expected_req_count)

        # Check sleep called as expected
        assert remove_web_retry_backoff_sleep.call_args_list == expected_sleep_calls

    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=_HTTP_CALL_TYPES,
        indirect=True,
    )
    @responses.activate
    def test_retry_handles_connection_error(
        self,
        caplog: LogCaptureFixture,
        http_func: Callable[..., Response],
        http_method: str,
        remove_web_retry_backoff_sleep: Optional[Mock],
        test_url: str,
    ) -> None:
        """Test that retry occurs when connection error occurs.

        Tests for all HTTP method types.
        """
        # Check retry sleep is patched
        assert remove_web_retry_backoff_sleep is not None

        # By not adding a URL to `responses` we can force a ConnectionError to be raised
        with pytest.raises(requests.ConnectionError), caplog.at_level(logging.DEBUG):
            http_func(url=test_url)

        # Check sleep called as expected
        assert remove_web_retry_backoff_sleep.call_args_list == [
            call(i) for i in _DEFAULT_BACKOFF_TIMES
        ]

        # Check logs assert this
        debug_logs = get_debug_logs(caplog)
        for retry_count, backoff in [
            (i + 1, backoff) for i, backoff in enumerate(_DEFAULT_BACKOFF_TIMES)
        ]:
            backoff_message_regex = (
                rf"Connection error \(.*\) for {http_method}:{test_url}/?;"
                rf" will retry in {backoff} seconds \(attempt {retry_count}\)\."
            )
            assert re.search(backoff_message_regex, debug_logs, re.DOTALL)

    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=_HTTP_CALL_TYPES,
        indirect=True,
    )
    @responses.activate
    def test_retry_handles_empty_connection_error(
        self,
        caplog: LogCaptureFixture,
        http_func: Callable[..., Response],
        http_method: str,
        remove_web_retry_backoff_sleep: Optional[Mock],
        test_url: str,
    ) -> None:
        """Test that retry occurs if empty ConnectionError occurs.

        That is one that doesn't contain the request object.

        Rests for all HTTP method types.
        """
        # Check retry sleep is patched
        assert remove_web_retry_backoff_sleep is not None

        # To force an empty ConnectionError, we need to set the exception manually
        # on the response
        exc = requests.ConnectionError("err_str")
        responses.add(
            method=http_method,
            url=test_url,
            body=exc,
        )

        with pytest.raises(
            requests.ConnectionError, match=re.escape("err_str")
        ), caplog.at_level(logging.DEBUG):
            http_func(url=test_url)

        # Check sleep called as expected
        assert remove_web_retry_backoff_sleep.call_args_list == [
            call(i) for i in _DEFAULT_BACKOFF_TIMES
        ]

        # Check logs assert this
        debug_logs = get_debug_logs(caplog)
        for retry_count, backoff in [
            (i + 1, backoff) for i, backoff in enumerate(_DEFAULT_BACKOFF_TIMES)
        ]:
            backoff_message_regex = (
                rf"Connection error \({str(exc)}\);"
                rf" will retry in {backoff} seconds \(attempt {retry_count}\)\."
            )
            assert re.search(backoff_message_regex, debug_logs, re.DOTALL)

    @pytest.mark.parametrize(
        "backoff_factor",
        (None, 2 * web_utils._DEFAULT_BACKOFF_FACTOR),
        ids=["default_backoff", "non_default_backoff"],
    )
    @pytest.mark.parametrize(
        "max_retries",
        (None, 2 * web_utils._DEFAULT_MAX_RETRIES),
        ids=["default_max_retries", "non_default_max_retries"],
    )
    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=(
            "head",
            "get",
            "post",
            "put",
            "patch",
            "delete",
        ),
        indirect=True,
    )
    @responses.activate
    def test_decorator_works_with_diff_args(
        self,
        backoff_factor: Optional[int],
        http_method: str,
        max_retries: Optional[int],
        remove_web_retry_backoff_sleep: Optional[Mock],
        test_url: str,
    ) -> None:
        """Tests that the retry decorator works when used with args or brackets."""
        # Check retry sleep is patched
        assert remove_web_retry_backoff_sleep is not None

        class _DecKwargs(TypedDict):
            # This class is just to assuage mypy when we pass it to the decorator later
            max_retries: NotRequired[int]
            backoff_factor: NotRequired[int]

        dec_kwargs: _DecKwargs = {}
        if max_retries:
            dec_kwargs["max_retries"] = max_retries
        if backoff_factor:
            dec_kwargs["backoff_factor"] = backoff_factor

        # Decorator call will either be empty brackets, "max_retries=X",
        # "backoff_factor=X", or "max_retries=X, backoff_factor=X"
        @_auto_retry_request(**dec_kwargs)
        def request_(*args: Any, **kwargs: Any) -> Response:
            return requests.request(*args, **kwargs)

        with pytest.raises(requests.ConnectionError):
            request_(method=http_method, url=test_url)

        # Check sleep/retry was as expected
        used_backoff_factor = (
            backoff_factor if backoff_factor else web_utils._DEFAULT_BACKOFF_FACTOR
        )
        used_max_retries = (
            max_retries if max_retries else web_utils._DEFAULT_MAX_RETRIES
        )
        expected_backoff_times = tuple(
            used_backoff_factor * (2**i) for i in range(used_max_retries)
        )
        assert remove_web_retry_backoff_sleep.call_args_list == [
            call(i) for i in expected_backoff_times
        ]

    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=_HTTP_CALL_TYPES,
        indirect=True,
    )
    @pytest.mark.parametrize(
        "provide_timeout", (True, False), ids=lambda x: f"provide_timeout={x}"
    )
    @responses.activate
    def test_default_timeout_set_if_missing(
        self,
        caplog: LogCaptureFixture,
        http_func: Callable[..., Response],
        http_method: str,
        provide_timeout: bool,
        test_url: str,
    ) -> None:
        """Tests that default timeout handling works correctly."""
        if provide_timeout:
            # Ensure a distinct value for the set timeout
            timeout = web_utils._DEFAULT_TIMEOUT / 2
        else:
            timeout = web_utils._DEFAULT_TIMEOUT

        responses.add(
            method=http_method,
            url=test_url,
            match=[
                responses.matchers.request_kwargs_matcher({"timeout": timeout}),
            ],
        )

        with caplog.at_level(logging.DEBUG):
            if provide_timeout:
                http_func(url=test_url, timeout=timeout)
            else:
                http_func(url=test_url)

        # Check logs
        debug_logs = get_debug_logs(caplog)
        expected_log = (
            f"No request timeout provided,"
            f" setting to default timeout ({web_utils._DEFAULT_TIMEOUT}s)"
        )
        if provide_timeout:
            assert expected_log not in debug_logs
        else:
            assert expected_log in debug_logs


@unit_test
class TestASyncWebUtils:
    """Tests for asynchronous web interactions."""

    @fixture
    def http_func(
        self, http_func_details: str, http_method: str
    ) -> Callable[..., Awaitable[httpx.Response]]:
        """The actual HTTP function to use.

        This will either be the syntactic sugar version (i.e. `.async_get(...)`)
        or the direct version (i.e. `.async_request(method="GET", ...)` depending
        on the form of http_func_details.
        """
        if http_func_details.startswith("request"):
            return partial(web_utils.async_request, method=http_method)
        else:
            return cast(
                Callable[..., Awaitable[httpx.Response]],
                getattr(web_utils, f"async_{http_method.lower()}"),
            )

    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=_HTTP_CALL_TYPES,
        indirect=True,
    )
    @pytest.mark.parametrize("status_code", _HTTP_STATUS_CODES)
    async def test_retry_handles_http_status_codes(
        self,
        http_func: Callable[..., Awaitable[httpx.Response]],
        http_method: str,
        httpx_mock: HTTPXMock,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        status_code: int,
        test_url: str,
    ) -> None:
        """Test that auto retry occurs for expected HTTP status codes.

        Checks for all status codes and HTTP method combinations.
        """
        # Check retry sleep is patched
        assert remove_async_web_retry_backoff_sleep is not None

        httpx_mock.add_response(
            url=test_url,
            status_code=status_code,
            method=http_method,
        )

        resp = await http_func(url=test_url)

        assert resp.status_code == status_code
        assert resp.request.method == http_method

        # Construct expected values depending on if retryable
        if status_code in web_utils._RETRY_STATUS_CODES:
            expected_req_count = 1 + web_utils._DEFAULT_MAX_RETRIES
            expected_sleep_calls = [call(i) for i in _DEFAULT_BACKOFF_TIMES]
        else:
            expected_req_count = 1
            expected_sleep_calls = []

        # Check if retries occurred as expected
        assert len(httpx_mock.get_requests(url=test_url)) == expected_req_count

        # Check sleep called as expected
        assert (
            remove_async_web_retry_backoff_sleep.await_args_list == expected_sleep_calls
        )

    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=_HTTP_CALL_TYPES,
        indirect=True,
    )
    async def test_retry_handles_connection_error(
        self,
        caplog: LogCaptureFixture,
        http_func: Callable[..., Awaitable[httpx.Response]],
        http_method: str,
        httpx_mock: HTTPXMock,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        test_url: str,
    ) -> None:
        """Test that retry occurs when connection error occurs.

        Tests for all HTTP method types.
        """
        # Check retry sleep is patched
        assert remove_async_web_retry_backoff_sleep is not None

        httpx_mock.add_exception(httpx.ConnectError("err_str"), url=test_url)

        with pytest.raises(httpx.ConnectError), caplog.at_level(logging.DEBUG):
            await http_func(url=test_url)

        # Check sleep called as expected
        assert remove_async_web_retry_backoff_sleep.await_args_list == [
            call(i) for i in _DEFAULT_BACKOFF_TIMES
        ]

        # Check logs assert this
        debug_logs = get_debug_logs(caplog)
        for retry_count, backoff in [
            (i + 1, backoff) for i, backoff in enumerate(_DEFAULT_BACKOFF_TIMES)
        ]:
            backoff_message_regex = (
                rf"Connection error \(.*\) for {http_method}:{test_url}/?;"
                rf" will retry in {backoff} seconds \(attempt {retry_count}\)\."
            )
            assert re.search(backoff_message_regex, debug_logs, re.DOTALL)

    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=_HTTP_CALL_TYPES,
        indirect=True,
    )
    async def test_retry_handles_empty_connection_error(
        self,
        caplog: LogCaptureFixture,
        http_func: Callable[..., Awaitable[httpx.Response]],
        http_method: str,
        httpx_mock: HTTPXMock,
        mocker: MockerFixture,
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        test_url: str,
    ) -> None:
        """Test that retry occurs if empty ConnectionError occurs.

        That is one that doesn't contain the request object.

        Rests for all HTTP method types.
        """
        # Check retry sleep is patched
        assert remove_async_web_retry_backoff_sleep is not None

        # To force an empty ConnectError, we need to set the exception manually
        # on the response. We need to set the request property getter to raise an
        # exceptions as this signifies that the request has not been set.
        def _property_mock_getter_exception(*args: Any) -> None:
            if len(args) != 0:
                # Setter called
                return None
            else:
                # Getter called
                raise RuntimeError

        exc = httpx.ConnectError("err_str")
        mocker.patch.object(
            type(exc),
            "request",
            PropertyMock(side_effect=_property_mock_getter_exception),
        )
        httpx_mock.add_exception(exc, url=test_url)

        with pytest.raises(
            httpx.ConnectError, match=re.escape("err_str")
        ), caplog.at_level(logging.DEBUG):
            await http_func(url=test_url)

        # Check sleep called as expected
        assert remove_async_web_retry_backoff_sleep.await_args_list == [
            call(i) for i in _DEFAULT_BACKOFF_TIMES
        ]

        # Check logs assert this
        debug_logs = get_debug_logs(caplog)
        for retry_count, backoff in [
            (i + 1, backoff) for i, backoff in enumerate(_DEFAULT_BACKOFF_TIMES)
        ]:
            backoff_message_regex = (
                rf"Connection error \({str(exc)}\);"
                rf" will retry in {backoff} seconds \(attempt {retry_count}\)\."
            )
            assert re.search(backoff_message_regex, debug_logs, re.DOTALL)

    @pytest.mark.parametrize(
        "backoff_factor",
        (None, 2 * web_utils._DEFAULT_BACKOFF_FACTOR),
        ids=["default_backoff", "non_default_backoff"],
    )
    @pytest.mark.parametrize(
        "max_retries",
        (None, 2 * web_utils._DEFAULT_MAX_RETRIES),
        ids=["default_max_retries", "non_default_max_retries"],
    )
    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=(
            "head",
            "get",
            "post",
            "put",
            "patch",
            "delete",
        ),
        indirect=True,
    )
    async def test_decorator_works_with_diff_args(
        self,
        backoff_factor: Optional[int],
        http_method: str,
        httpx_mock: HTTPXMock,
        max_retries: Optional[int],
        remove_async_web_retry_backoff_sleep: Optional[AsyncMock],
        test_url: str,
    ) -> None:
        """Tests that the retry decorator works when used with args or brackets."""
        # Check retry sleep is patched
        assert remove_async_web_retry_backoff_sleep is not None

        class _DecKwargs(TypedDict):
            # This class is just to assuage mypy when we pass it to the decorator later
            max_retries: NotRequired[int]
            backoff_factor: NotRequired[int]

        dec_kwargs: _DecKwargs = {}
        if max_retries:
            dec_kwargs["max_retries"] = max_retries
        if backoff_factor:
            dec_kwargs["backoff_factor"] = backoff_factor

        # Decorator call will either be empty brackets, "max_retries=X",
        # "backoff_factor=X", or "max_retries=X, backoff_factor=X"
        @_auto_retry_request(**dec_kwargs)
        async def request_(*args: Any, **kwargs: Any) -> httpx.Response:
            async with httpx.AsyncClient() as client:
                return await client.request(*args, **kwargs)

        httpx_mock.add_exception(httpx.ConnectError("err_str"))

        with pytest.raises(httpx.ConnectError):
            await request_(method=http_method, url=test_url)

        # Check sleep/retry was as expected
        used_backoff_factor = (
            backoff_factor if backoff_factor else web_utils._DEFAULT_BACKOFF_FACTOR
        )
        used_max_retries = (
            max_retries if max_retries else web_utils._DEFAULT_MAX_RETRIES
        )
        expected_backoff_times = tuple(
            used_backoff_factor * (2**i) for i in range(used_max_retries)
        )
        assert remove_async_web_retry_backoff_sleep.await_args_list == [
            call(i) for i in expected_backoff_times
        ]

    @pytest.mark.parametrize(
        argnames="http_func_details",
        argvalues=_HTTP_CALL_TYPES,
        indirect=True,
    )
    @pytest.mark.parametrize(
        "provide_timeout", (True, False), ids=lambda x: f"provide_timeout={x}"
    )
    async def test_default_timeout_set_if_missing(
        self,
        caplog: LogCaptureFixture,
        http_func: Callable[..., Awaitable[httpx.Response]],
        http_method: str,
        httpx_mock: HTTPXMock,
        provide_timeout: bool,
        test_url: str,
    ) -> None:
        """Tests that default timeout handling works correctly."""
        if provide_timeout:
            # Ensure a distinct value for the set timeout
            timeout = web_utils._DEFAULT_TIMEOUT / 2
        else:
            timeout = web_utils._DEFAULT_TIMEOUT

        def check_timeout_callback(request: httpx.Request) -> Optional[httpx.Response]:
            # The form of the HTTPX timeout information is a dict of different timeouts
            if provide_timeout:
                # All timeouts will be the same
                assert request.extensions["timeout"] == {
                    k: timeout for k in ("connect", "read", "write", "pool")
                }
            else:
                # Default timeouts are only set on the "connect" and "pool" attributes
                assert request.extensions["timeout"] == {
                    "connect": timeout,
                    "read": None,
                    "write": None,
                    "pool": timeout,
                }
            return httpx.Response(status_code=200)

        httpx_mock.add_callback(
            callback=check_timeout_callback,
            method=http_method,
            url=test_url,
        )

        with caplog.at_level(logging.DEBUG):
            if provide_timeout:
                await http_func(url=test_url, timeout=timeout)
            else:
                await http_func(url=test_url)

        # Check logs
        debug_logs = get_debug_logs(caplog)
        expected_log = (
            f"No request timeout provided,"
            f" setting to default timeout ({web_utils._DEFAULT_TIMEOUT}s)"
        )
        if provide_timeout:
            assert expected_log not in debug_logs
        else:
            assert expected_log in debug_logs



"""Tests `bitfount/federated/encryption.py`."""
import base64
from pathlib import Path
import re
from typing import Optional
from unittest.mock import create_autospec

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.hazmat.primitives.hashes import HashAlgorithm
import pytest
from pytest import fixture
from pytest_mock import MockerFixture

from bitfount.federated.encryption import (
    _AES_NONCE_SIZE_BYTES,
    _AESEncryption,
    _calc_max_RSA_message_size,
    _read_file,
    _RSAEncryption,
)
from bitfount.federated.exceptions import DecryptError
from tests.bitfount import TEST_SECURITY_FILES
from tests.utils.helper import (
    PRIVATE_KEY,
    PRIVATE_SSH_KEY,
    PUBLIC_KEY,
    PUBLIC_SSH_KEY,
    unit_test,
)

MAX_BYTE_VALUE = 255
PUBLIC_KEY_PATH = TEST_SECURITY_FILES / "test_public.testkey"


@unit_test
class TestEncryptionUtils:
    """Tests encryption utility functions."""

    def test_read_file(self) -> None:
        """Tests `read_file` returns bytes."""
        contents = _read_file(PUBLIC_KEY_PATH)
        assert contents is not None
        assert isinstance(contents, bytes)

    @pytest.mark.parametrize(
        argnames=("rsa_key_size_bits", "oaep_hash_algo", "expected_max_size"),
        argvalues=(
            # Taken from https://crypto.stackexchange.com/a/42100
            (1024, hashes.SHA1(), 86),
            (1024, hashes.SHA224(), 70),
            (1024, hashes.SHA256(), 62),
            (1024, hashes.SHA384(), 30),
            (2048, hashes.SHA1(), 214),
            (2048, hashes.SHA224(), 198),
            (2048, hashes.SHA256(), 190),
            (2048, hashes.SHA384(), 158),
            (2048, hashes.SHA512(), 126),
            (3072, hashes.SHA1(), 342),
            (3072, hashes.SHA224(), 326),
            (3072, hashes.SHA256(), 318),
            (3072, hashes.SHA384(), 286),
            (3072, hashes.SHA512(), 254),
            (4096, hashes.SHA1(), 470),
            (4096, hashes.SHA224(), 454),
            (4096, hashes.SHA256(), 446),
            (4096, hashes.SHA384(), 414),
            (4096, hashes.SHA512(), 382),
        ),
    )
    @pytest.mark.parametrize("private_or_public_key", (True, False))
    def test__calc_max_RSA_message_size(
        self,
        expected_max_size: int,
        mocker: MockerFixture,
        oaep_hash_algo: HashAlgorithm,
        private_or_public_key: bool,
        rsa_key_size_bits: int,
    ) -> None:
        """Tests max RSA message size calculator against known values."""
        # Create key mock with target size
        if private_or_public_key:
            mock_key = create_autospec(RSAPrivateKey, instance=True)
        else:
            mock_key = create_autospec(RSAPublicKey, instance=True)
        mock_key.key_size = rsa_key_size_bits

        # Patch out retrieval of OAEP digest size
        mock__get_encryption_padding = mocker.patch.object(
            _RSAEncryption,
            "_get_encryption_padding",
        )
        mock__get_encryption_padding.return_value._algorithm = oaep_hash_algo

        assert _calc_max_RSA_message_size(mock_key) == expected_max_size


@unit_test
class TestRSAEncryption:
    """Tests for RSAEncryption class."""

    @fixture
    def private_key_file(self) -> Path:
        """Returns private key file."""
        return TEST_SECURITY_FILES / "test_private.testkey"

    @fixture
    def private_key(self, private_key_file: Path) -> RSAPrivateKey:
        """Loaded private key from private key file."""
        return _RSAEncryption.load_private_key(private_key_file)

    @fixture
    def public_key_file(self) -> Path:
        """Returns public key file."""
        return TEST_SECURITY_FILES / "test_public.testkey"

    @fixture
    def public_key(self, public_key_file: Path) -> RSAPublicKey:
        """Loaded public key from public key file."""
        return _RSAEncryption.load_public_key(public_key_file)

    @fixture
    def public_key_size_bytes(self, public_key: RSAPublicKey) -> int:
        """Key size for public key fixture."""
        return (public_key.key_size + 7) // 8

    @fixture
    def signature_file(self) -> Path:
        """Returns signature file."""
        return TEST_SECURITY_FILES / "test_signature.sign"

    @fixture
    def private_key_bytes(self) -> bytes:
        """Returns private key bytes."""
        return PRIVATE_KEY.encode()

    @fixture
    def public_key_bytes(self) -> bytes:
        """Returns public key bytes."""
        return PUBLIC_KEY.encode()

    @fixture
    def private_ssh_key_file(self) -> Path:
        """Returns private ssh key file."""
        return TEST_SECURITY_FILES / "test_ssh_key_rsa.sshtestkey"

    @fixture
    def public_ssh_key_file(self) -> Path:
        """Returns public ssh key file."""
        return TEST_SECURITY_FILES / "test_ssh_key_rsa.pub.sshtestkey"

    @fixture
    def private_ssh_key_bytes(self) -> bytes:
        """Returns private ssh key bytes."""
        return PRIVATE_SSH_KEY.encode()

    @fixture
    def public_ssh_key_bytes(self) -> bytes:
        """Returns public ssh key bytes."""
        return PUBLIC_SSH_KEY.encode()

    def test_generate_key_pair_produces_rsa_keys(self) -> None:
        """Checks that an RSAPublicKey and an RSAPrivateKey are generated."""
        private_key, public_key = _RSAEncryption.generate_key_pair()

        assert private_key.key_size >= 2048

        assert isinstance(private_key, RSAPrivateKey)
        assert isinstance(public_key, RSAPublicKey)

    def test_load_public_key(
        self, public_key_bytes: bytes, public_key_file: Path
    ) -> None:
        """Tests loading of PEM public key from file and bytes."""
        public_key = _RSAEncryption.load_public_key(public_key_file)
        assert public_key is not None
        assert isinstance(public_key, RSAPublicKey)

        public_key = _RSAEncryption.load_public_key(public_key_bytes)
        assert public_key is not None
        assert isinstance(public_key, RSAPublicKey)

    def test_load_public_ssh_key(
        self, public_ssh_key_bytes: bytes, public_ssh_key_file: Path
    ) -> None:
        """Tests loading of SSH public key from file and bytes."""
        public_key = _RSAEncryption.load_public_key(public_ssh_key_file)
        assert public_key is not None
        assert isinstance(public_key, RSAPublicKey)

        public_key = _RSAEncryption.load_public_key(public_ssh_key_bytes)
        assert public_key is not None
        assert isinstance(public_key, RSAPublicKey)

    def test_serialized_and_deserialized_public_key_is_same_key(self) -> None:
        """Checks serialization and deserialization for PEM-format public keys."""
        _, public_key = _RSAEncryption.generate_key_pair()

        reloaded_key = _RSAEncryption.load_public_key(
            _RSAEncryption.serialize_public_key(public_key)
        )

        assert public_key.public_numbers() == reloaded_key.public_numbers()

    def test_load_private_key(
        self, private_key_bytes: bytes, private_key_file: Path
    ) -> None:
        """Tests loading of PEM private key from file and bytes."""
        private_key = _RSAEncryption.load_private_key(private_key_file)
        assert private_key is not None
        assert isinstance(private_key, RSAPrivateKey)

        private_key = _RSAEncryption.load_private_key(private_key_bytes)
        assert private_key is not None
        assert isinstance(private_key, RSAPrivateKey)

    def test_load_private_ssh_key(
        self, private_ssh_key_bytes: bytes, private_ssh_key_file: Path
    ) -> None:
        """Tests loading of SSH private key from file and bytes."""
        private_key = _RSAEncryption.load_private_key(private_ssh_key_file)
        assert private_key is not None
        assert isinstance(private_key, RSAPrivateKey)

        private_key = _RSAEncryption.load_private_key(private_ssh_key_bytes)
        assert private_key is not None
        assert isinstance(private_key, RSAPrivateKey)

    def test_serialized_and_deserialized_private_key_is_same_key(self) -> None:
        """Checks serialization and deserialization for PEM-format private keys."""
        private_key, _ = _RSAEncryption.generate_key_pair()

        reloaded_key = _RSAEncryption.load_private_key(
            _RSAEncryption.serialize_private_key(private_key)
        )

        assert private_key.private_numbers() == reloaded_key.private_numbers()

    def test_sign_message(self, private_key: RSAPrivateKey) -> None:
        """Tests message signing produces byte signatures."""
        message = b"Test*123"
        signature = _RSAEncryption.sign_message(private_key, message)
        assert signature is not None
        assert isinstance(signature, bytes)

    def test_verify_javascript_signed_signature(
        self, public_key: RSAPublicKey, signature_file: Path
    ) -> None:
        """Checks message signature verification using external values.

        Uses signatures and message from javascript tests.
        """
        message = "Test*123".encode("ascii")
        with open(signature_file, "r") as f:
            signature = f.readline().encode("ascii")
            signature = base64.b64decode(signature)
        result = _RSAEncryption.verify_signature(public_key, signature, message)
        assert result

    def test_verify_signature(
        self,
        private_key: RSAPrivateKey,
        public_key: RSAPublicKey,
    ) -> None:
        """Checks message signature verification works."""
        message = b"Hello world"
        signature = _RSAEncryption.sign_message(private_key, message)
        result = _RSAEncryption.verify_signature(public_key, signature, message)
        assert result

        # Try again with a different message but same signature
        message = b"Goodbye world"
        result = _RSAEncryption.verify_signature(public_key, signature, message)
        assert result is False

    def test_rsa_correct_decryption(
        self,
        private_key: RSAPrivateKey,
        public_key: RSAPublicKey,
    ) -> None:
        """Tests encryption-decryption cycle in RSAEncryption."""
        message = b"Hello world"
        ciphertext = _RSAEncryption.encrypt(message, public_key)

        # Assert Decryption is correct
        plaintext = _RSAEncryption.decrypt(ciphertext, private_key)
        assert message == plaintext

    def test_rsa_failed_decryption(
        self,
        private_key: RSAPrivateKey,
        public_key: RSAPublicKey,
    ) -> None:
        """Checks decryption raises exception if incorrect ciphertext."""
        message = b"Hello world"
        ciphertext = _RSAEncryption.encrypt(message, public_key)

        ciphertext = bytearray(ciphertext)
        ciphertext[1] = MAX_BYTE_VALUE - ciphertext[1]
        ciphertext = bytes(ciphertext)

        with pytest.raises(
            DecryptError, match=re.escape("Unable to decrypt RSA-encrypted message.")
        ):
            _RSAEncryption.decrypt(ciphertext, private_key)

    def test_rsa_uses_hybrid_encryption_for_too_large_messages(
        self,
        mocker: MockerFixture,
        private_key: RSAPrivateKey,
        public_key: RSAPublicKey,
    ) -> None:
        """Checks hybrid encryption used for too large messages."""
        # Wrap AES encryption so we can check it was used
        wrapped_aes = mocker.patch(
            "bitfount.federated.encryption._AESEncryption", wraps=_AESEncryption
        )

        # Make a message _just_ over the maximum size
        message = b"a" * (_calc_max_RSA_message_size(public_key) + 1)

        encrypted_message = _RSAEncryption.encrypt(message, public_key)
        decrypted_message = _RSAEncryption.decrypt(encrypted_message, private_key)

        assert decrypted_message == message

        # Check AES methods used
        wrapped_aes.generate_key.assert_called_once()
        wrapped_aes.encrypt.assert_called_once()
        wrapped_aes.decrypt.assert_called_once()

    @pytest.mark.parametrize(
        argnames=("invalidate_index_shift", "expected_error_msg"),
        argvalues=(
            pytest.param(
                None,
                "Unable to decrypt AES key in hybrid RSA-encrypted message.",
                id="aes_key_invalid",
            ),
            pytest.param(
                0,
                "Unable to decrypt ciphertext in hybrid RSA-encrypted message.",
                id="aes_nonce_invalid",
            ),
            pytest.param(
                _AES_NONCE_SIZE_BYTES,
                "Unable to decrypt ciphertext in hybrid RSA-encrypted message.",
                id="aes_ciphertext_invalid",
            ),
        ),
    )
    def test_hybrid_decryption_fails_with_incorrect_ciphertext(
        self,
        expected_error_msg: str,
        invalidate_index_shift: Optional[int],
        private_key: RSAPrivateKey,
        public_key: RSAPublicKey,
        public_key_size_bytes: int,
    ) -> None:
        """Checks hybrid encryption fails if ciphertext incorrect.

        Due to a quirk of lazy_fixture (https://github.com/TvoroG/pytest-lazy-fixture/issues/24)  # noqa: B950
        we can't use that in the parameterization so instead use whether a shift
        is present or not to determine which base index we're working from.
        """
        # Make a message _just_ over the maximum size
        message = b"a" * (_calc_max_RSA_message_size(public_key) + 1)

        encrypted_message = _RSAEncryption.encrypt(message, public_key)

        # Modify message to invalid state
        if invalidate_index_shift is None:
            invalidate_index = 0
        else:
            invalidate_index = public_key_size_bytes + invalidate_index_shift
        encrypted_message = bytearray(encrypted_message)
        encrypted_message[invalidate_index] = (
            MAX_BYTE_VALUE - encrypted_message[invalidate_index]
        )
        encrypted_message = bytes(encrypted_message)

        with pytest.raises(DecryptError, match=re.escape(expected_error_msg)):
            _RSAEncryption.decrypt(encrypted_message, private_key)


@unit_test
class TestAESEncryption:
    """Tests AESEncryption class."""

    @fixture
    def key(self) -> bytes:
        """An AESEncryption key to use in tests."""
        return _AESEncryption.generate_key()

    @fixture
    def message(self) -> bytes:
        """Plaintext message to encrypt."""
        return b"Hello world"

    def test_aes_correct_encryption_decryption(
        self, key: bytes, message: bytes
    ) -> None:
        """Tests encryption-decryption cycle in AESEncryption."""
        ciphertext, nonce = _AESEncryption.encrypt(key, message)
        plaintext = _AESEncryption.decrypt(key, nonce, ciphertext)
        assert plaintext == message

    def test_aes_wrong_nonce_decryption(self, key: bytes, message: bytes) -> None:
        """Checks decryption raises error if incorrect nonce."""
        ciphertext, nonce = _AESEncryption.encrypt(key, message)
        wrong_nonce = bytes((MAX_BYTE_VALUE - i) for i in nonce)
        with pytest.raises(DecryptError):
            _AESEncryption.decrypt(key, wrong_nonce, ciphertext)

    def test_aes_failed_decryption(self, key: bytes, message: bytes) -> None:
        """Checks decryption raises error if incorrect ciphertext."""
        ciphertext, nonce = _AESEncryption.encrypt(key, message)
        ciphertext = bytearray(ciphertext)
        ciphertext[1] = MAX_BYTE_VALUE - ciphertext[1]
        ciphertext = bytes(ciphertext)
        with pytest.raises(
            DecryptError, match=re.escape("Unable to decrypt ciphertext")
        ):
            _AESEncryption.decrypt(key, nonce, ciphertext)



"""Tests for datafactory.py."""
from unittest.mock import create_autospec

from _pytest.monkeypatch import MonkeyPatch
import pytest

from bitfount.config import _BASIC_ENGINE, _PYTORCH_ENGINE
from bitfount.data.datafactory import _BasicDataFactory, _get_default_data_factory
from bitfount.data.dataloaders import BitfountDataLoader
from bitfount.data.datasets import _BaseBitfountDataset
from bitfount.exceptions import BitfountEngineError
from tests.utils.helper import unit_test


@unit_test
class TestDefaultDataFactoryLoading:
    """Tests the default factory loading."""

    def test_load_default_data_factory(self, monkeypatch: MonkeyPatch) -> None:
        """Test that the default data factory can load."""
        # Set envvar value
        monkeypatch.setattr("bitfount.data.datafactory.BITFOUNT_ENGINE", _BASIC_ENGINE)
        df = _get_default_data_factory()
        assert isinstance(df, _BasicDataFactory)

    def test_load_default_data_factory_fail_on_import_error(
        self, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests handling if the imported library throws an ImportError."""
        # Try with PyTorch backend, even if can't be imported here
        monkeypatch.setattr(
            "bitfount.data.datafactory.BITFOUNT_ENGINE", _PYTORCH_ENGINE
        )
        try:
            monkeypatch.delattr(
                "bitfount.backends.pytorch.data.datafactory._PyTorchDataFactory"
            )
        except ImportError:
            pass

        with pytest.raises(BitfountEngineError, match="pytorch"):
            _get_default_data_factory()

    def test_load_default_data_factory_fails_unknown_engine(
        self, monkeypatch: MonkeyPatch
    ) -> None:
        """Tests that if the engine type is unknown, an error is raised."""
        monkeypatch.setattr(
            "bitfount.data.datafactory.BITFOUNT_ENGINE", "NOT_AN_ENGINE"
        )

        with pytest.raises(
            BitfountEngineError, match="Unable to load engine NOT_AN_ENGINE."
        ):
            _get_default_data_factory()

    def test_create_dataloader(self) -> None:
        """Tests create_dataloader returns BitfountDataLoader."""
        data_factory = _BasicDataFactory()
        dataset = create_autospec(_BaseBitfountDataset)
        dataloader = data_factory.create_dataloader(dataset)
        assert isinstance(dataloader, BitfountDataLoader)



"""Helper functions."""
from __future__ import annotations

import re
from typing import TYPE_CHECKING, Iterable, List, Mapping, Optional, Sequence, Union

from bitfount.config import (
    _DEVELOPMENT_ENVIRONMENT,
    _PRODUCTION_ENVIRONMENT,
    _STAGING_ENVIRONMENT,
    _get_environment,
)
from bitfount.data.exceptions import BitfountSchemaError
from bitfount.data.schema import BitfountSchema
from bitfount.federated.aggregators.aggregator import Aggregator
from bitfount.federated.aggregators.base import _BaseAggregatorFactory
from bitfount.federated.aggregators.secure import SecureAggregator
from bitfount.federated.exceptions import PodNameError
from bitfount.federated.logging import _get_federated_logger
from bitfount.federated.secure import SecureShare
from bitfount.federated.transport.config import (
    _DEV_MESSAGE_SERVICE_PORT,
    _DEV_MESSAGE_SERVICE_TLS,
    _DEV_MESSAGE_SERVICE_URL,
    _STAGING_MESSAGE_SERVICE_URL,
    MessageServiceConfig,
)
from bitfount.federated.transport.message_service import _MessageService
from bitfount.federated.transport.pod_transport import _PodMailbox
from bitfount.hub.helper import get_pod_schema
from bitfount.hub.types import _DEV_IDP_URL, _PRODUCTION_IDP_URL, _STAGING_IDP_URL
from bitfount.runners.config_schemas import POD_NAME_REGEX, USERNAME_REGEX

if TYPE_CHECKING:
    from bitfount.hub.api import BitfountHub
    from bitfount.hub.authentication_flow import BitfountSession

logger = _get_federated_logger(__name__)


def _create_aggregator(
    secure_aggregation: bool,
    weights: Optional[Mapping[str, Union[int, float]]] = None,
) -> _BaseAggregatorFactory:
    """Creates aggregator for Federated Averaging.

    Args:
        model: The model used in aggregation.
        secure_aggregation: Boolean denoting whether aggregator should be secure.
        weights: Per-pod update weighting to use when aggregating updates,
            or None if equal weighting required.

    Raises:
        TypeError: If model is not compatible with Federated Averaging.

    Returns:
        The aggregator to be used.
    """
    if secure_aggregation and weights:
        # TODO: [BIT-1486] Remove this constraint
        raise NotImplementedError("SecureAggregation does not support update weighting")

    if secure_aggregation:
        sec_share = SecureShare()
        return SecureAggregator(secure_share=sec_share)
    return Aggregator(weights=weights)


def _get_idp_url() -> str:
    """Helper function for defining idp url based on environment."""
    environment = _get_environment()
    if environment == _STAGING_ENVIRONMENT:
        idp_url = _STAGING_IDP_URL
    elif environment == _DEVELOPMENT_ENVIRONMENT:
        idp_url = _DEV_IDP_URL
    elif environment == _PRODUCTION_ENVIRONMENT:
        idp_url = _PRODUCTION_IDP_URL
    return idp_url


def _create_message_service(
    session: BitfountSession, ms_config: Optional[MessageServiceConfig] = None
) -> _MessageService:
    """Helper function to create MessageService object.

    Args:
        session (BitfountSession): bitfount session
        ms_config (Optional[MessageServiceConfig], optional): message service config.
            Defaults to None.

    Returns:
        MessageService object
    """
    if ms_config is None:
        ms_config = MessageServiceConfig()

        environment = _get_environment()
        if environment == _STAGING_ENVIRONMENT:
            ms_config.url = _STAGING_MESSAGE_SERVICE_URL
        elif environment == _DEVELOPMENT_ENVIRONMENT:
            ms_config.url = _DEV_MESSAGE_SERVICE_URL
            ms_config.port = _DEV_MESSAGE_SERVICE_PORT
            ms_config.tls = _DEV_MESSAGE_SERVICE_TLS

    if ms_config.use_local_storage:
        logger.warning(
            "Messages will contain local file references. "
            + "Ensure all pods have access to your local file system. "
            + "Otherwise your task will hang.",
        )

    return _MessageService(session, ms_config)


async def _create_and_connect_pod_mailbox(
    pod_name: str,
    session: BitfountSession,
    ms_config: Optional[MessageServiceConfig] = None,
) -> _PodMailbox:
    """Creates pod mailbox and connects it to the message service.

    Args:
        pod_name: Name of pod.
        session: Bitfount session.
        ms_config: Optional. Message service config, defaults to None.

    Returns:
        The created pod mailbox.
    """
    message_service = _create_message_service(session, ms_config)
    mailbox = await _PodMailbox.connect_pod(
        pod_name=pod_name, message_service=message_service
    )
    return mailbox


def _check_and_update_pod_ids(
    pod_identifiers: Iterable[str], hub: BitfountHub
) -> List[str]:
    """Add username from hub to pod identifiers if not already provided."""
    # TODO: [BIT-991] check if pod id exists
    pod_id_regex = re.compile(USERNAME_REGEX.pattern + "/" + POD_NAME_REGEX.pattern)
    updated_pod_ids = []
    for pod_id in pod_identifiers:
        if pod_id_regex.fullmatch(pod_id):
            updated_pod_ids.append(pod_id)
        elif "/" not in pod_id:
            if POD_NAME_REGEX.fullmatch(pod_id):
                updated_pod_ids.append(f"{hub.username}/{pod_id}")
            else:
                raise PodNameError(
                    f"Invalid Pod name : {pod_id}."
                    "Pod name must consist of lower case alphanumeric "
                    "characters optionally seperated by '-'."
                )
        else:
            raise PodNameError(
                f"Invalid Pod name : {pod_id}. "
                "Pod name must be a have the format "
                "<username>/<pod_name>. Pod name must consist "
                "of lower case alphanumeric "
                "characters optionally seperated by '-'."
            )

    return updated_pod_ids


def combine_pod_schemas(pod_names: Sequence[str]) -> BitfountSchema:
    """Combines the schemas of multiple pods into a single schema.

    Simply aggregates all the tables in the schemas of the pods into a single
    schema.

    Args:
        pod_names: The names of the pods whose schemas will be combined.

    Returns:
        The combined schema.

    Raises:
        BitfountSchemaError: If the schemas of the pods are incompatible.
    """
    combined_schema = BitfountSchema()
    for pod_name in pod_names:
        pod_schema: BitfountSchema = get_pod_schema(pod_name)
        combined_schema.tables.extend(pod_schema.tables)

    if len(combined_schema.table_names) != len(set(combined_schema.table_names)):
        raise BitfountSchemaError("Multiple tables found with the same name.")

    return combined_schema



"""Tests for LightGBM models."""
import os
from pathlib import Path

import lightgbm
import pytest
from pytest_mock import MockerFixture
import sqlalchemy

from bitfount.backends.lightgbm.models.models import (
    LGBMRandomForestClassifier,
    LGBMRandomForestRegressor,
)
from bitfount.data.datasources.base_source import BaseSource
from bitfount.data.datasources.database_source import DatabaseSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.data.types import SchemaOverrideMapping
from bitfount.data.utils import DatabaseConnection
from bitfount.models.base_models import MAIN_MODEL_REGISTRY
from bitfount.schemas.utils import bf_dump, bf_load
from tests.bitfount.models.test_models import SERIALIZED_MODEL_NAME, assert_vars_equal
from tests.utils.helper import (
    assert_results,
    backend_test,
    create_datasource,
    create_datastructure,
    create_query_datastructure,
    create_schema,
    integration_test,
    unit_test,
)


@pytest.fixture
def datastructure() -> DataStructure:
    """Fixture for datastructure."""
    return create_datastructure()


@pytest.fixture
def query_datastructure() -> DataStructure:
    """Fixture for datastructure containing query."""
    return create_query_datastructure()


@pytest.fixture
def datasource() -> BaseSource:
    """Fixture for datasource."""
    return create_datasource(classification=True)


@pytest.fixture
def schema() -> BitfountSchema:
    """Fixture for datastructure with schema."""
    return create_schema(classification=True)


@backend_test
class TestLGBMRandomForest:
    """Test LGBMRandomForest model classes."""

    @integration_test
    def test_classification(
        self,
        datasource: BaseSource,
        datastructure: DataStructure,
        schema: BitfountSchema,
    ) -> None:
        """Tests LGBMRandomForestClassifier training."""
        random_forest = LGBMRandomForestClassifier(
            datastructure=datastructure,
            schema=schema,
            n_estimators=10,
            early_stopping_rounds=2,
            verbose=-1,
        )
        random_forest.fit(datasource)
        assert_results(model=random_forest)

    @integration_test
    def test_regression(
        self, datasource: BaseSource, datastructure: DataStructure
    ) -> None:
        """Tests LGBMRandomForestRegressor training."""
        random_forest = LGBMRandomForestRegressor(
            datastructure=datastructure,
            schema=create_schema(classification=False),
            verbose=-1,
        )
        random_forest.fit(datasource)
        assert_results(model=random_forest)

    @integration_test
    def test_serialization(
        self,
        datasource: BaseSource,
        datastructure: DataStructure,
        schema: BitfountSchema,
        tmp_path: Path,
    ) -> None:
        """Tests serialize() and deserialize() methods."""
        random_forest = LGBMRandomForestClassifier(
            datastructure=datastructure,
            schema=schema,
            n_estimators=10,
            early_stopping_rounds=10,
            verbose=-1,
        )

        random_forest.fit(datasource)
        random_forest.serialize(str(tmp_path / SERIALIZED_MODEL_NAME))
        assert os.path.exists(tmp_path / SERIALIZED_MODEL_NAME)
        rf_model = LGBMRandomForestClassifier(
            datastructure=datastructure,
            schema=schema,
            n_estimators=10,
            early_stopping_rounds=10,
            verbose=-1,
        )
        rf_model.fit(datasource)
        rf_model.deserialize(str(tmp_path / SERIALIZED_MODEL_NAME))
        rf_model.evaluate(random_forest.test_set)

    @unit_test
    def test_fit_called_with_query_datastructure(
        self,
        mocker: MockerFixture,
    ) -> None:
        """Tests model is fit with a query datastructure."""
        datasource = create_datasource(classification=True)
        schema = BitfountSchema(datasource, table_name="TABLE")
        schema_override: SchemaOverrideMapping = {
            "categorical": [
                {"M": {"False": 0, "True": 1}},
                {"TARGET": {"0": 0, "1": 1}},
            ],
            "continuous": ["A"],
        }
        query = "SELECT * from TABLE"
        ds = DataStructure(
            target="TARGET", query=query, schema_types_override=schema_override
        )
        random_forest = LGBMRandomForestClassifier(
            datastructure=ds,
            schema=schema,
            verbose=-1,
        )

        mocker.patch.object(lightgbm, "train", return_value=None)
        random_forest.fit(datasource)
        assert random_forest.databunch is not None
        assert random_forest.n_classes == 2

    @integration_test
    def test_fit_called_with_db(
        self,
        db_session: sqlalchemy.engine.base.Engine,
        mocker: MockerFixture,
    ) -> None:
        """Tests model is fit with a query datastructure."""
        datastructure = DataStructure(target="TARGET", table="dummy_data")
        db_datasource = DatabaseSource(
            DatabaseConnection(db_session, table_names=["dummy_data"])
        )
        schema = BitfountSchema(
            db_datasource,
            table_name="dummy_data",
            force_stypes={"dummy_data": {"categorical": ["TARGET"]}},
        )
        random_forest = LGBMRandomForestClassifier(
            datastructure=datastructure,
            schema=schema,
            verbose=-1,
        )

        mocker.patch.object(lightgbm, "train", return_value=None)
        random_forest.fit(db_datasource)
        assert random_forest.databunch is not None
        assert random_forest.n_classes == 2

    @unit_test
    def test_fit_called_with_table_datastructure(
        self,
        datasource: BaseSource,
        datastructure: DataStructure,
        mocker: MockerFixture,
        schema: BitfountSchema,
    ) -> None:
        """Tests model is fit with a table datastructure."""
        random_forest = LGBMRandomForestClassifier(
            datastructure=datastructure,
            schema=schema,
            n_estimators=10,
            early_stopping_rounds=2,
            verbose=-1,
        )
        mocker.patch.object(lightgbm, "train", return_value=None)

        random_forest.fit(datasource)
        assert random_forest.databunch is not None
        assert random_forest.n_classes == 2

    @unit_test
    def test_evaluate_no_test_dl_error(self, datastructure: DataStructure) -> None:
        """Tests that evaluate raises error with no test_dl."""
        random_forest = LGBMRandomForestRegressor(
            datastructure=datastructure, schema=BitfountSchema(), verbose=-1
        )
        with pytest.raises(ValueError):
            random_forest.evaluate()

    @unit_test
    def test_fit_no_validation_dl(
        self,
        datasource: BaseSource,
        datastructure: DataStructure,
        schema: BitfountSchema,
    ) -> None:
        """Tests that evaluate called without test data raises error."""
        random_forest = LGBMRandomForestClassifier(
            datastructure=datastructure,
            schema=schema,
            n_estimators=10,
            early_stopping_rounds=2,
            verbose=-1,
        )
        datasource.load_data()
        random_forest._add_datasource_to_schema(datasource)
        random_forest._set_dataloaders()
        random_forest.validation_dl = None
        train_df, val_df = random_forest._create_dataset()
        assert val_df is None


@backend_test
@unit_test
class TestMarshmallowSerialization:
    """Test Marshmallow Serialization for LightGBM models."""

    def test_rf_classifier_serialization(
        self, datastructure: DataStructure, schema: BitfountSchema
    ) -> None:
        """Tests serialization with LGBMRandomForestClassifier."""
        model = LGBMRandomForestClassifier(datastructure=datastructure, schema=schema)
        serialized_model = bf_dump(model)
        deserialized_model = bf_load(serialized_model, MAIN_MODEL_REGISTRY)
        assert_vars_equal(vars(model), vars(deserialized_model))

    def test_rf_regressor_serialization(self, datastructure: DataStructure) -> None:
        """Tests serialization with LGBMRandomForestRegressor."""
        model = LGBMRandomForestRegressor(
            datastructure=datastructure, schema=BitfountSchema()
        )
        serialized_model = bf_dump(model)
        deserialized_model = bf_load(serialized_model, MAIN_MODEL_REGISTRY)
        assert_vars_equal(vars(model), vars(deserialized_model))



"""Symmetric and asymmetric encryption functions."""
from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING, Final, Optional, Tuple, Union, cast

from cryptography.exceptions import InvalidSignature, InvalidTag
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.asymmetric.rsa import (
    RSAPrivateKey,
    RSAPrivateKeyWithSerialization,
    RSAPublicKey,
)
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
)

from bitfount.federated.exceptions import DecryptError
from bitfount.federated.logging import _get_federated_logger

if TYPE_CHECKING:
    from cryptography.hazmat.primitives.asymmetric.padding import MGF
    from cryptography.hazmat.primitives.hashes import HashAlgorithm


logger = _get_federated_logger(__name__)

# RSA Constants
_RSA_PUBLIC_EXPONENT: Final[int] = 65537
_RSA_KEY_SIZE_BITS: Final[int] = 4096
# Salt length is padding.PSS.MAX_LENGTH, but we manually set it to ensure
# compatibility across platforms.
_RSA_SIGN_SALT_LENGTH: Final[int] = 20

# AES Constants
# The AES key must provide at least as much security as the RSA key as they are
# both used in hybrid encryption (when the message is too large for RSA encryption
# alone).
#
# This NIST-based table allows us to find values which will enable that:
#   https://www.keylength.com/en/4/
_AES_KEY_SIZE_BITS: Final[int] = 256
_AES_NONCE_SIZE_BYTES: Final[int] = 12


def _read_file(file_to_read: Path) -> bytes:
    """Reads given file and returns contents as a byte string."""
    with open(file_to_read, "rb") as f:
        contents = f.read()
    return contents


def _calc_max_RSA_message_size(rsa_key: Union[RSAPrivateKey, RSAPublicKey]) -> int:
    """Calculates the maximum message size that can be encrypted.

    Calculates the maximum message size, in bytes, that can be encrypted with
    the supplied key.

    https://www.rfc-editor.org/rfc/rfc8017#section-7.1.1
    """
    # This is the approach used internally to cryptography
    # Rounded up byte-size from bit-size
    rsa_key_size_bytes = (rsa_key.key_size + 7) // 8

    # Get the expected padding settings and access the algorithm to get the
    # expected output size.
    oaep_hash_output_size_bytes = (
        _RSAEncryption._get_encryption_padding()._algorithm.digest_size
    )

    # Equation from: https://www.rfc-editor.org/rfc/rfc8017#section-7.1.1
    max_size = rsa_key_size_bytes - (2 * oaep_hash_output_size_bytes) - 2
    return max_size


class _RSAEncryption:
    """Class of functions for dealing with RSA asymmetric encryption."""

    @staticmethod
    def generate_key_pair() -> Tuple[RSAPrivateKey, RSAPublicKey]:
        """Generates a new RSA key pair."""
        logger.debug("Generating RSA key pair")
        # Key size is 4096 bits which means we can only encrypt up to 4096 bits of data
        private_key = rsa.generate_private_key(
            public_exponent=_RSA_PUBLIC_EXPONENT,
            key_size=_RSA_KEY_SIZE_BITS,
            backend=default_backend(),
        )
        return private_key, private_key.public_key()

    @staticmethod
    def load_private_key(private_key: Union[bytes, Path]) -> RSAPrivateKey:
        """Loads a private key either from a byte string or file path."""
        logger.debug("Attempting to load private key")
        if isinstance(private_key, Path):
            logger.debug(f"Loading private key from path: {private_key}")
            private_key = _read_file(private_key)

        # Try loading from PEM format first
        try:
            logger.debug("Attempting to load private key using PEM Format...")
            loaded_private_key = serialization.load_pem_private_key(
                private_key, password=None, backend=default_backend()
            )

        except ValueError as ve:
            # Otherwise try SSH format
            try:
                logger.debug(
                    "Loading private key using PEM format failed,"
                    " trying to load using SSH format..."
                )
                loaded_private_key = serialization.load_ssh_private_key(
                    private_key, password=None, backend=default_backend()
                )
            except ValueError:
                # If both fail, raise the original error
                raise ve

        logger.debug("Loaded private key")
        return cast(RSAPrivateKey, loaded_private_key)

    @staticmethod
    def serialize_private_key(private_key: RSAPrivateKeyWithSerialization) -> bytes:
        """Serializes a private key to bytes."""
        return private_key.private_bytes(
            Encoding.PEM, PrivateFormat.TraditionalOpenSSL, NoEncryption()
        )

    @staticmethod
    def load_public_key(public_key: Union[bytes, Path]) -> RSAPublicKey:
        """Loads a public key either from a byte string or file path."""
        logger.debug("Attempting to load public key")
        if isinstance(public_key, Path):
            logger.debug(f"Loading public key from path: {public_key}")
            public_key = _read_file(public_key)

        # Try loading from PEM format first
        try:
            logger.debug("Attempting to load public key using PEM Format...")
            loaded_public_key = serialization.load_pem_public_key(
                public_key, backend=default_backend()
            )
        except ValueError as ve:
            # Otherwise try SSH format
            try:
                logger.debug(
                    "Loading public key using PEM format failed, "
                    "trying to load using SSH format..."
                )
                loaded_public_key = serialization.load_ssh_public_key(
                    public_key, backend=default_backend()
                )
            except ValueError:
                # If both fail, raise the original error
                raise ve

        logger.debug("Loaded public key")
        return cast(RSAPublicKey, loaded_public_key)

    @staticmethod
    def serialize_public_key(public_key: RSAPublicKey) -> bytes:
        """Serialize an RSAPublicKey to bytes."""
        return public_key.public_bytes(Encoding.PEM, PublicFormat.PKCS1)

    @staticmethod
    def _get_hashing_algorithm() -> HashAlgorithm:
        """Retrieves a new instance representing the chosen hash algorithm."""
        return hashes.SHA256()

    @staticmethod
    def _get_mask_gen_function() -> MGF:
        """Retrieves a new instance representing the chosen mask generator function."""
        return padding.MGF1(_RSAEncryption._get_hashing_algorithm())

    @staticmethod
    def _get_signature_padding() -> padding.PSS:
        """Retrieves a new instance representing the padding algo for signatures."""
        return padding.PSS(
            mgf=_RSAEncryption._get_mask_gen_function(),
            salt_length=_RSA_SIGN_SALT_LENGTH,
        )

    @staticmethod
    def _get_encryption_padding() -> padding.OAEP:
        """Retrieves a new instance representing the padding algo for encryption."""
        return padding.OAEP(
            mgf=_RSAEncryption._get_mask_gen_function(),
            algorithm=_RSAEncryption._get_hashing_algorithm(),
            label=None,
        )

    @staticmethod
    def sign_message(private_key: RSAPrivateKey, message: bytes) -> bytes:
        """Cryptographically signs a message.

        Signs provided `message` with provided `private_key` and returns signature.
        """
        signature = private_key.sign(
            message,
            padding=_RSAEncryption._get_signature_padding(),
            algorithm=_RSAEncryption._get_hashing_algorithm(),
        )

        return signature

    @staticmethod
    def verify_signature(
        public_key: RSAPublicKey, signature: bytes, message: bytes
    ) -> bool:
        """Verifies that decrypting `signature` with `public_key` === `message`."""
        try:
            public_key.verify(
                signature,
                message,
                padding=_RSAEncryption._get_signature_padding(),
                algorithm=_RSAEncryption._get_hashing_algorithm(),
            )
        except InvalidSignature:
            return False

        logger.debug("Signature verified")
        return True

    @staticmethod
    def encrypt(message: bytes, public_key: RSAPublicKey) -> bytes:
        """Encrypts plaintext.

        Encrypts provided `message` with `public_key` and returns ciphertext.
        """
        # If message is small enough to be fully RSA encrypted then do that.
        if len(message) <= _calc_max_RSA_message_size(public_key):
            ciphertext = public_key.encrypt(
                message,
                _RSAEncryption._get_encryption_padding(),
            )
        # Otherwise, use hybrid encryption.
        else:
            logger.debug(
                "Message too large for RSA encryption, using hybrid encryption."
            )

            # Encrypt message with AES key
            aes_key = _AESEncryption.generate_key()
            aes_ciphertext, aes_nonce = _AESEncryption.encrypt(aes_key, message)

            # Encrypt AES key using RSA encryption
            encrypted_aes_key = public_key.encrypt(
                aes_key,
                _RSAEncryption._get_encryption_padding(),
            )

            # Combine
            ciphertext = encrypted_aes_key + aes_nonce + aes_ciphertext
        return ciphertext

    @staticmethod
    def decrypt(
        ciphertext: bytes,
        private_key: RSAPrivateKey,
    ) -> bytes:
        """Decrypts ciphertext.

        Decrypts provided `ciphertext` with `private_key` and returns plaintext.

        Raises:
            DecryptError: if decryption fails.
        """
        # This is the approach used internally to cryptography
        # Rounded up byte-size from bit-size
        rsa_encrypted_size = (private_key.key_size + 7) // 8

        # If message is the expected RSA size then must have been purely RSA
        # encrypted. Decrypt directly.
        if len(ciphertext) <= rsa_encrypted_size:
            try:
                plaintext = private_key.decrypt(
                    ciphertext,
                    _RSAEncryption._get_encryption_padding(),
                )
            except ValueError:
                raise DecryptError("Unable to decrypt RSA-encrypted message.")
        # Otherwise, it must have been encrypted using hybrid encryption.
        else:
            logger.debug(
                "Message too large for RSA encryption, using hybrid decryption."
            )

            # Split received message into relevant parts.
            encrypted_aes_key, aes_payload = (
                ciphertext[:rsa_encrypted_size],
                ciphertext[rsa_encrypted_size:],
            )
            aes_nonce, aes_ciphertext = (
                aes_payload[:_AES_NONCE_SIZE_BYTES],
                aes_payload[_AES_NONCE_SIZE_BYTES:],
            )

            # Decrypt AES key
            try:
                aes_key = private_key.decrypt(
                    encrypted_aes_key,
                    _RSAEncryption._get_encryption_padding(),
                )
            except ValueError:
                raise DecryptError(
                    "Unable to decrypt AES key in hybrid RSA-encrypted message."
                )

            # Decrypt message body
            try:
                plaintext = _AESEncryption.decrypt(aes_key, aes_nonce, aes_ciphertext)
            except DecryptError:
                raise DecryptError(
                    "Unable to decrypt ciphertext in hybrid RSA-encrypted message."
                )

        return plaintext


class _AESEncryption:
    """Class of functions for dealing with AES symmetric encryption."""

    @staticmethod
    def generate_key() -> bytes:
        """Generates a symmetric encryption key.

        Generates symmetric key using the GCM algorithm (Galois Counter Mode).

        (More secure than CBC (Cipher Block Chaining)).
        """
        key = AESGCM.generate_key(
            bit_length=_AES_KEY_SIZE_BITS
        )  # 128 bits is sufficiently secure
        return key

    @staticmethod
    def encrypt(
        key: bytes, plaintext: bytes, associated_data: Optional[bytes] = None
    ) -> Tuple[bytes, bytes]:
        """Encrypts plaintext.

        Encrypts `plaintext` using `key` and a randomly generated `nonce`.
        If `associated_data` is provided, it is authenticated.
        """
        aesgcm = AESGCM(key)
        nonce = os.urandom(_AES_NONCE_SIZE_BYTES)  # 12 bytes
        ciphertext = aesgcm.encrypt(nonce, plaintext, associated_data)

        return ciphertext, nonce

    @staticmethod
    def decrypt(
        key: bytes,
        nonce: bytes,
        ciphertext: bytes,
        associated_data: Optional[bytes] = None,
    ) -> bytes:
        """Decrypts ciphertext.

        Decrypts `ciphertext` using `key`, `nonce` and `associated_data` if present.

        If `associated_data` is provided, this must be the same associated data
        used in encryption.

        ***
        NONCE MUST ONLY BE USED ONCE FOR A GIVEN KEY (SAME NONCE AS USED FOR
        ENCRYPTION)
        ***

        Raises:
            DecryptError: if decryption fails.
        """
        aesgcm = AESGCM(key)

        try:
            plaintext = aesgcm.decrypt(nonce, ciphertext, associated_data)
        except InvalidTag:
            raise DecryptError("Unable to decrypt ciphertext")

        return plaintext



"""Tests for modeller-side OIDC handling."""
import asyncio
from asyncio import Task
import base64
from datetime import datetime, timedelta, timezone
import functools
import hashlib
import logging
import re
import secrets
from typing import Dict, List, Literal, Optional, Type, cast
from unittest.mock import Mock, call, create_autospec
from urllib.parse import urlencode

from _pytest.capture import CaptureFixture
from _pytest.logging import LogCaptureFixture
from aiohttp import web
from aiohttp.pytest_plugin import AiohttpClient
from aiohttp.test_utils import TestClient
from aiohttp.web import Application
from aiohttp.web_runner import TCPSite
import pytest
from pytest import fixture
from pytest_mock import MockerFixture
from requests import HTTPError
from requests.exceptions import InvalidJSONError
import responses
from responses import matchers

from bitfount import web_utils
from bitfount.federated.transport.identity_verification import _BITFOUNT_MODELLER_PORT
from bitfount.federated.transport.identity_verification.oidc import (
    _CALLBACK_ROUTE,
    _get_urlsafe_hash,
    _OIDCAuthFlowChallengeHandler,
    _OIDCDeviceCodeHandler,
    _OIDCWebEndpoint,
)
from bitfount.federated.transport.modeller_transport import _ModellerMailbox
from bitfount.federated.transport.types import (
    _DeviceCodeDetailsPair,
    _ModellerDeviceCodeDetails,
    _OIDCAuthEndpointResponse,
    _OIDCAuthFlowResponse,
    _OIDCClientID,
    _PodDeviceCodeDetails,
)
from bitfount.hub.authentication_flow import AuthEnvironmentError, _AuthEnv
from bitfount.hub.types import _DeviceCodeRequestDict, _DeviceCodeResponseJSON
from tests.utils.helper import get_critical_logs, get_info_logs, unit_test
from tests.utils.mocks import AwaitableMock


@unit_test
class TestOIDCWebEndpoint:
    """Tests for the OIDC web endpoint."""

    @fixture
    def oidc_web_endpoint(self) -> _OIDCWebEndpoint:
        """A web endpoint for OIDC authentication."""
        return _OIDCWebEndpoint()

    @fixture
    def mock_webbrowser(self, mocker: MockerFixture) -> Mock:
        """Mock webbrowser import."""
        return mocker.patch(
            "bitfount.federated.transport.identity_verification.oidc.webbrowser",
            autospec=True,
        )

    @fixture
    def authorize_urls(self) -> List[str]:
        """List of URLs for authorize calls."""
        return ["first_url", "second_url", "third_url"]

    @fixture
    def pod_identifiers(self) -> List[str]:
        """Pod identifiers."""
        return ["user1/pod1", "user2/pod2", "user3/pod3"]

    @fixture
    def states(self, pod_identifiers: List[str]) -> Dict[str, str]:
        """Randomly generated states for API calls."""
        return {pod_id: secrets.token_hex(nbytes=32) for pod_id in pod_identifiers}

    @fixture
    def auth_codes(self) -> List[str]:
        """Fake auth codes as though generated from /authorize."""
        return [f"auth_code_{i}" for i in range(3)]

    @fixture
    def callback_urls(self, auth_codes: List[str], states: Dict[str, str]) -> List[str]:
        """Callback URLs as though generated from /authorize."""
        return [
            f'{_CALLBACK_ROUTE}?{urlencode({"code": auth_codes[i], "state": states[pod_id]})}'  # noqa: B950
            for i, pod_id in enumerate(states)
        ]

    @fixture
    async def client(
        self, aiohttp_client: AiohttpClient, oidc_web_endpoint: _OIDCWebEndpoint
    ) -> TestClient:
        """Test client for AioHTTP interactions."""
        app = Application()
        app.add_routes([web.get(_CALLBACK_ROUTE, oidc_web_endpoint.process_callback)])
        return await aiohttp_client(app)

    def test_start_processing(
        self,
        authorize_urls: List[str],
        caplog: LogCaptureFixture,
        mock_webbrowser: Mock,
        oidc_web_endpoint: _OIDCWebEndpoint,
        states: Dict[str, str],
    ) -> None:
        """Test start processing opens the initial url."""
        # Initialise endpoint
        oidc_web_endpoint.initialise(authorize_urls, states)

        with caplog.at_level(logging.INFO):
            oidc_web_endpoint.start_processing()

        # Check first URL opened
        mock_webbrowser.open.assert_called_once_with(authorize_urls[0])
        # Check logged out first URL
        info_logs = get_info_logs(caplog)
        assert (
            f"Attempting to open browser. Running a headless client? "
            f"You'll need to open this link in a browser: {authorize_urls[0]}"
            in info_logs
        )
        # Check first URL dropped from iterator
        assert list(oidc_web_endpoint._urls_iter) == authorize_urls[1:]

    def test_start_processing_throws_error_if_not_initialized(
        self,
        caplog: LogCaptureFixture,
        mock_webbrowser: Mock,
        oidc_web_endpoint: _OIDCWebEndpoint,
    ) -> None:
        """Test start processing throws exception if not initialised."""
        expected_log_message = (
            "OIDC callback endpoint has not been initialised. "
            "Unable to complete OIDC authentication."
        )

        with pytest.raises(RuntimeError, match=re.escape(expected_log_message)):
            oidc_web_endpoint.start_processing()

        # Check logged out error
        critical_logs = get_critical_logs(caplog)
        assert expected_log_message in critical_logs
        # Check webbrowser not opened
        mock_webbrowser.open.assert_not_called()

    def test_initialise_sets_initialised_variable(
        self,
        authorize_urls: List[str],
        oidc_web_endpoint: _OIDCWebEndpoint,
        states: Dict[str, str],
    ) -> None:
        """Tests that initialise() sets initialised variable."""
        assert oidc_web_endpoint._initialised is False
        oidc_web_endpoint.initialise(authorize_urls, states)
        assert oidc_web_endpoint._initialised is True

    async def test_process_callback_loop(
        self,
        auth_codes: List[str],
        authorize_urls: List[str],
        callback_urls: List[str],
        client: TestClient,
        mock_webbrowser: Mock,
        oidc_web_endpoint: _OIDCWebEndpoint,
        states: Dict[str, str],
    ) -> None:
        """Tests process_callback() successfully redirects to all authorizations."""
        oidc_web_endpoint.initialise(authorize_urls, states)

        # These requests would normally be made by the user's browser after being
        # redirected after making a /authorize API call (which happens in the
        # `start_processing()` call).
        oidc_web_endpoint.start_processing()
        # First OIDC authorize response (i.e. first callback URL)
        first_authorize_redirect = await client.get(
            callback_urls[0], allow_redirects=False
        )
        # First redirect takes us to the second /authorize URL as the first is
        # opened by the user.
        assert first_authorize_redirect.headers["location"] == authorize_urls[1]
        assert not oidc_web_endpoint.all_done.is_set()

        # Second OIDC authorize response (i.e. second callback URL)
        second_authorize_redirect = await client.get(
            callback_urls[1], allow_redirects=False
        )
        # Second redirect takes us to the third /authorize URL.
        assert second_authorize_redirect.headers["location"] == authorize_urls[2]
        assert not oidc_web_endpoint.all_done.is_set()

        # Third callback URL should take us to the success page
        success_page = await client.get(callback_urls[2], allow_redirects=False)
        assert str(success_page.url.path) == _CALLBACK_ROUTE
        page_text = await success_page.text()
        assert (
            page_text == "You've now proven your identity to all pods involved"
            " in the task. You can close this tab."
        )
        # Should be all done now (including empty URLs iterator)
        assert oidc_web_endpoint.all_done.is_set()
        with pytest.raises(StopIteration):
            next(oidc_web_endpoint._urls_iter)

        # Check states match
        assert oidc_web_endpoint._id_to_state == states
        # Check responses
        responses = oidc_web_endpoint._responses
        assert responses == {
            pod_id: _OIDCAuthEndpointResponse(
                auth_codes[i],
                states[pod_id],
            )
            for i, pod_id in enumerate(states)
        }

    async def test_process_callback_fails_if_not_initialized(
        self, caplog: LogCaptureFixture, oidc_web_endpoint: _OIDCWebEndpoint
    ) -> None:
        """Test process_callback() raises exception if not initialized."""
        expected_log_message = (
            "OIDC callback endpoint has not been initialised. "
            "Unable to complete OIDC authentication."
        )

        with pytest.raises(RuntimeError, match=re.escape(expected_log_message)):
            await oidc_web_endpoint.process_callback(request=Mock())

        # Check logged out error
        critical_logs = get_critical_logs(caplog)
        assert expected_log_message in critical_logs

    async def test_get_responses_waits_for_completion(
        self, mocker: MockerFixture, oidc_web_endpoint: _OIDCWebEndpoint
    ) -> None:
        """Tests that get_responses() waits for all_done to be set."""
        # Mock out responses
        mock_responses = mocker.patch.object(
            oidc_web_endpoint, "_responses", autospec=True
        )

        # Mock out all_done event
        mock_all_done = mocker.patch.object(
            oidc_web_endpoint, "all_done", autospec=True
        )

        responses = await oidc_web_endpoint.get_responses()

        # Check waited on all_done event
        mock_all_done.wait.assert_awaited_once()
        # Check returned correctly
        assert responses == mock_responses


@pytest.mark.parametrize(argnames="initial_encoding", argvalues=("utf-8", "ascii"))
@unit_test
def test__get_urlsafe_hash(initial_encoding: str) -> None:
    """Test _get_urlsafe_hash() produces expected hash."""
    test_string = "Hello, world!"

    # Create expected hash
    encoded_s: bytes = test_string.encode(initial_encoding)
    sha256_hash: bytes = hashlib.sha256(encoded_s).digest()
    b64_hash: bytes = base64.urlsafe_b64encode(sha256_hash)
    b64_hash_str: str = b64_hash.decode("utf-8").replace("=", "")

    assert _get_urlsafe_hash(test_string) == b64_hash_str


@fixture
def auth_domain() -> str:
    """Fake auth domain."""
    return "fake_auth_domain"


@fixture
def scopes() -> str:
    """Fake oauth scopes."""
    return "fake scopes for auth0"


@fixture
def audience() -> str:
    """Fake oauth audience."""
    return "fake_audience"


@fixture
def client_id() -> str:
    """Fake client ID."""
    return "fake_client_id"


@fixture
def fake_auth_env(auth_domain: str, client_id: str) -> _AuthEnv:
    """Fake _AuthEnv instance for tests."""
    return _AuthEnv(
        name="fake_auth_env",
        auth_domain=auth_domain,
        client_id=client_id,
    )


@fixture
def patch_get_auth_environment(fake_auth_env: _AuthEnv, mocker: MockerFixture) -> Mock:
    """Patch out the return value of _get_auth_environment() to our fake env."""
    mock_get_auth_environment = mocker.patch(
        "bitfount.federated.transport.identity_verification.oidc._get_auth_environment",  # noqa: B950
        autospec=True,
        return_value=fake_auth_env,
    )
    return mock_get_auth_environment


@fixture
def mock_modeller_mailbox() -> Mock:
    """Mock _ModellerMailbox instance."""
    mock_mailbox: Mock = create_autospec(_ModellerMailbox, instance=True)
    return mock_mailbox


@fixture
def num_pods() -> int:
    """The number of pods to replicate in handle tests."""
    return 3


@fixture
def pod_identifiers(num_pods: int) -> List[str]:
    """Pod identifiers."""
    return [f"user{i}/pod{i}" for i in range(num_pods)]


@fixture
def received_oidc_details(
    client_id: str, pod_identifiers: List[str]
) -> Dict[str, _OIDCClientID]:
    """Fake received OIDC details as would be received from pods."""
    return {pod_id: _OIDCClientID(client_id) for pod_id in pod_identifiers}


@unit_test
class TestOIDCAuthFlowChallengeHandler:
    """Tests for _OIDCAuthFlowChallengeHandler."""

    @fixture
    def oidc_auth_flow_handler(
        self, audience: str, auth_domain: str, scopes: str
    ) -> _OIDCAuthFlowChallengeHandler:
        """OIDC Authorisation Code Flow challenge handler."""
        return _OIDCAuthFlowChallengeHandler(auth_domain, scopes, audience)

    @fixture
    def code_verifiers(self, num_pods: int) -> List[str]:
        """Pregenerated code verifiers for /authorize requests."""
        return [secrets.token_urlsafe(nbytes=60) for _ in range(num_pods)]

    @fixture
    def code_challenges(self, code_verifiers: List[str]) -> List[str]:
        """Code challenges for the pregenerated code verifiers."""
        return [
            _get_urlsafe_hash(cv, initial_encoding="ascii") for cv in code_verifiers
        ]

    @fixture
    def states(self, num_pods: int) -> List[str]:
        """Pregenerated states for /authorize requests."""
        return [secrets.token_hex(nbytes=32) for _ in range(num_pods)]

    @fixture
    def states_dict(
        self, pod_identifiers: List[str], states: List[str]
    ) -> Dict[str, str]:
        """Dictionary of pod IDs to the pregenerated states."""
        return {pod_id: state for pod_id, state in zip(pod_identifiers, states)}

    @fixture
    def expected_authorize_urls(
        self,
        audience: str,
        client_id: str,
        code_challenges: List[str],
        num_pods: int,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
        scopes: str,
        states: List[str],
    ) -> List[str]:
        """Expected /authorize URLs (with pregenerated params)."""
        expected_params = [
            urlencode(
                {
                    "audience": audience,
                    "scope": scopes,
                    "response_type": "code",
                    "client_id": client_id,
                    "state": states[i],
                    "redirect_uri": oidc_auth_flow_handler._redirect_uri,
                    "code_challenge_method": "S256",
                    "code_challenge": code_challenges[i],
                }
            )
            for i in range(num_pods)
        ]
        return [
            f"{oidc_auth_flow_handler._authorize_endpoint}?{params}"
            for params in expected_params
        ]

    @fixture
    def auth_codes(self, num_pods: int) -> List[str]:
        """Fake auth codes to return from web endpoint."""
        return [f"auth_code_{i}" for i in range(num_pods)]

    @fixture
    def final_oidc_details(
        self,
        auth_codes: List[str],
        code_verifiers: List[str],
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
        pod_identifiers: List[str],
    ) -> Dict[str, _OIDCAuthFlowResponse]:
        """Final OIDC details to send back to pods."""
        return {
            pod_id: _OIDCAuthFlowResponse(
                auth_codes[i], code_verifiers[i], oidc_auth_flow_handler._redirect_uri
            )
            for i, pod_id in enumerate(pod_identifiers)
        }

    @fixture
    def fake_endpoint_responses(
        self, auth_codes: List[str], pod_identifiers: List[str], states: List[str]
    ) -> Dict[str, _OIDCAuthEndpointResponse]:
        """Fake _OIDCWebEndpoint.get_responses() return value."""
        return {
            pod_id: _OIDCAuthEndpointResponse(
                auth_codes[i],
                states[i],
            )
            for i, pod_id in enumerate(pod_identifiers)
        }

    async def test_start_server(
        self,
        mocker: MockerFixture,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
    ) -> None:
        """Test start() correctly starts the server."""
        # Mock out site start-up method and init
        mock_tcp_site = create_autospec(TCPSite)
        mock_tcp_site_init = mocker.patch(
            "bitfount.federated.transport.identity_verification.oidc.TCPSite",
            autospec=True,
            return_value=mock_tcp_site,
        )

        # Check server not marked as started
        with pytest.raises(AttributeError):
            oidc_auth_flow_handler._server_start_task

        server_start_task = oidc_auth_flow_handler.start_server()
        await server_start_task

        # Check server now marked as started
        assert oidc_auth_flow_handler._server_start_task.done()
        # Check TCP site created and started
        mock_tcp_site_init.assert_called_once_with(
            oidc_auth_flow_handler._runner, "localhost", _BITFOUNT_MODELLER_PORT
        )
        mock_tcp_site.start.assert_awaited_once()

    async def test__start_times_out_after_time(
        self,
        caplog: LogCaptureFixture,
        mocker: MockerFixture,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
    ) -> None:
        """Tests that _start() will timeout if server start hangs."""
        # Mock out runner
        mock_runner = mocker.patch.object(
            oidc_auth_flow_handler, "_runner", autospec=True
        )

        # Mock out site start-up method and init
        mock_tcp_site = create_autospec(TCPSite)
        mock_tcp_site_init = mocker.patch(
            "bitfount.federated.transport.identity_verification.oidc.TCPSite",
            autospec=True,
            return_value=mock_tcp_site,
        )

        # Make wait_for raise timeout exception
        mocker.patch.object(
            asyncio, "wait_for", side_effect=TimeoutError("custom timeout")
        )

        with pytest.raises(TimeoutError, match="custom timeout"):
            await oidc_auth_flow_handler._start()

        # Check setup called and site created
        mock_runner.setup.assert_awaited_once()
        mock_tcp_site_init.assert_called_once()

        # Check logged exception
        critical_logs = get_critical_logs(caplog)
        assert (
            "Timeout reached whilst trying to bind OIDC web endpoint" in critical_logs
        )

    async def test__start_handles_address_in_use_error(
        self,
        caplog: LogCaptureFixture,
        mocker: MockerFixture,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
    ) -> None:
        """Tests that _start() handles address being unavailable."""
        # Mock out runner
        mock_runner = mocker.patch.object(
            oidc_auth_flow_handler, "_runner", autospec=True
        )

        # Mock out site start-up method and init
        mock_tcp_site = create_autospec(TCPSite)
        mock_tcp_site_init = mocker.patch(
            "bitfount.federated.transport.identity_verification.oidc.TCPSite",
            autospec=True,
            return_value=mock_tcp_site,
        )
        # Make site.start() raise error indicating address unavailable
        mock_tcp_site.start.side_effect = OSError()

        with pytest.raises(OSError):
            await oidc_auth_flow_handler._start()

        # Check setup called and site created
        mock_runner.setup.assert_awaited_once()
        mock_tcp_site_init.assert_called_once()

        # Check logged exception
        critical_logs = get_critical_logs(caplog)
        assert "Unable to bind OIDC web endpoint" in critical_logs

    def test__verify_client_ids_successful(
        self,
        client_id: str,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
        patch_get_auth_environment: Mock,
    ) -> None:
        """Test _verify_client_ids() passes correctly."""
        oidc_details = {f"user{i}/pod{i}": _OIDCClientID(client_id) for i in range(3)}

        # Should just pass happily
        oidc_auth_flow_handler._verify_client_ids(oidc_details)

    def test__verify_client_ids_fails_diff_auth_domain(
        self,
        auth_domain: str,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
        patch_get_auth_environment: Mock,
    ) -> None:
        """Test _verify_client_ids() fails if auth domains differ."""
        # Change _auth_domain on the handler to induce mismatch
        oidc_auth_flow_handler._auth_domain = "not_the_same"

        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Mismatch between setup auth environment and current: "
                f"expected {oidc_auth_flow_handler._auth_domain}, got {auth_domain}"
            ),
        ):
            oidc_auth_flow_handler._verify_client_ids({})

    def test__verify_client_ids_fails_diff_client_ids(
        self,
        client_id: str,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
        patch_get_auth_environment: Mock,
    ) -> None:
        """Test _verify_client_ids() fails if client IDs differ."""
        # Have one "correct" client_id
        oidc_details = {
            "user0/pod0": _OIDCClientID(client_id),
        }
        # Add two "incorrect" client_ids
        incorrect_oidc_details = {
            f"user{i}/pod{i}": _OIDCClientID(f"diff_client_id_{i}") for i in range(1, 3)
        }
        oidc_details.update(incorrect_oidc_details)

        # Create expected error message
        error_pods = "; ".join(
            [
                f"Pod ID = {pod_id}, Client ID = {details.client_id}"
                for pod_id, details in incorrect_oidc_details.items()
            ]
        )
        expected_error_message = (
            f"Authorisation environments do not match. Expected client_id "
            f'"{client_id}" but the following pods mismatched: {error_pods}'
        )

        with pytest.raises(
            AuthEnvironmentError, match=re.escape(expected_error_message)
        ):
            oidc_auth_flow_handler._verify_client_ids(oidc_details)

    async def test_handle(
        self,
        caplog: LogCaptureFixture,
        code_verifiers: List[str],
        expected_authorize_urls: List[str],
        fake_endpoint_responses: Dict[str, _OIDCAuthEndpointResponse],
        final_oidc_details: Dict[str, _OIDCAuthFlowResponse],
        mock_modeller_mailbox: Mock,
        mocker: MockerFixture,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
        patch_get_auth_environment: Mock,
        received_oidc_details: Dict[str, _OIDCClientID],
        states: List[str],
        states_dict: Dict[str, str],
    ) -> None:
        """Tests handle() correctly processes OIDC challenges."""
        # Mock out receiving of OIDC details
        mock_modeller_mailbox.get_oidc_client_ids.return_value = received_oidc_details

        # Mock out code_verifier generation
        mock_secrets = mocker.patch(
            "bitfount.federated.transport.identity_verification.oidc.secrets",
            autospec=True,
        )
        mock_secrets.token_urlsafe.side_effect = code_verifiers

        # Mock out state generation
        mock_secrets.token_hex.side_effect = states

        # Mock out runner and web endpoint
        mock_runner = mocker.patch.object(
            oidc_auth_flow_handler, "_runner", autospec=True
        )
        mock_endpoint = mocker.patch.object(
            oidc_auth_flow_handler, "_oidc_endpoint", autospec=True
        )
        mock_endpoint.get_responses.return_value = fake_endpoint_responses

        # Mock out server_start_task
        mock_server_start_task = AwaitableMock(spec=Task)
        mock_server_start_task.done.return_value = False
        oidc_auth_flow_handler._server_start_task = mock_server_start_task

        # Make call
        with caplog.at_level(logging.INFO):
            await oidc_auth_flow_handler.handle(mock_modeller_mailbox)

        # Check logs
        info_logs = get_info_logs(caplog)
        assert "Waiting for OIDC challenge handler to start" in info_logs

        # Check endpoint interactions
        mock_endpoint.initialise.assert_called_once_with(
            expected_authorize_urls,
            states_dict,
        )
        mock_endpoint.start_processing.assert_called_once()

        # Check server_start_task and runner interactions
        mock_server_start_task.done.assert_called_once()
        mock_server_start_task.assert_awaited_once()
        mock_server_start_task.result.assert_called_once()

        # Check correct details sent onwards
        mock_modeller_mailbox.send_oidc_auth_flow_responses.assert_awaited_once_with(
            final_oidc_details
        )

        # Check server is shutdown
        mock_runner.cleanup.assert_awaited()

    async def test_handle_fails_different_states(
        self,
        caplog: LogCaptureFixture,
        code_verifiers: List[str],
        expected_authorize_urls: List[str],
        fake_endpoint_responses: Dict[str, _OIDCAuthEndpointResponse],
        mock_modeller_mailbox: Mock,
        mocker: MockerFixture,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
        patch_get_auth_environment: Mock,
        pod_identifiers: List[str],
        received_oidc_details: Dict[str, _OIDCClientID],
        states: List[str],
        states_dict: Dict[str, str],
    ) -> None:
        """Tests handle() fails if sent states and received states mismatch."""
        # Mock out receiving of OIDC details
        mock_modeller_mailbox.get_oidc_client_ids.return_value = received_oidc_details

        # Mock out code_verifier generation
        mock_secrets = mocker.patch(
            "bitfount.federated.transport.identity_verification.oidc.secrets",
            autospec=True,
        )
        mock_secrets.token_urlsafe.side_effect = code_verifiers

        # Mock out state generation
        mock_secrets.token_hex.side_effect = states

        # Mock out runner and web endpoint
        mock_runner = mocker.patch.object(
            oidc_auth_flow_handler, "_runner", autospec=True
        )
        mock_endpoint = mocker.patch.object(
            oidc_auth_flow_handler, "_oidc_endpoint", autospec=True
        )
        # Modify states to make not match on one
        fake_endpoint_responses[pod_identifiers[-1]].state = "incorrect_state"
        mock_endpoint.get_responses.return_value = fake_endpoint_responses

        mock_server_start_task = AwaitableMock(spec=Task)
        mock_server_start_task.done.return_value = True
        oidc_auth_flow_handler._server_start_task = mock_server_start_task

        # Make call
        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Unable to validate response intended for {pod_identifiers[-1]}"
            ),
        ), caplog.at_level(logging.INFO):
            await oidc_auth_flow_handler.handle(mock_modeller_mailbox)

        # Check log DIDN'T wait
        info_logs = get_info_logs(caplog)
        assert "Waiting for OIDC challenge handler to start" not in info_logs

        # Check endpoint interactions
        mock_endpoint.initialise.assert_called_once_with(
            expected_authorize_urls,
            states_dict,
        )
        mock_endpoint.start_processing.assert_called_once()

        # Check server_start_task and runner interactions
        mock_server_start_task.done.assert_called_once()
        mock_server_start_task.result.assert_called_once()
        mock_server_start_task.assert_not_awaited()
        # Check server is still shutdown
        mock_runner.cleanup.assert_awaited()

    async def test_handle_fails_if_server_not_started(
        self,
        mock_modeller_mailbox: Mock,
        mocker: MockerFixture,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
    ) -> None:
        """Tests handle() fails if server not started."""
        # Mock out runner
        mock_runner = mocker.patch.object(
            oidc_auth_flow_handler, "_runner", autospec=True
        )

        with pytest.raises(
            RuntimeError,
            match=re.escape(
                "OIDC server has not been started; "
                "ensure start_server() has been called."
            ),
        ):
            await oidc_auth_flow_handler.handle(mock_modeller_mailbox)

        # Check server is still shutdown
        mock_runner.cleanup.assert_awaited()

    @pytest.mark.parametrize(
        argnames="server_start_exception_cls",
        argvalues=(TimeoutError, OSError, Exception),
    )
    async def test_handle_reraises_exception_from_start_task(
        self,
        code_verifiers: List[str],
        mock_modeller_mailbox: Mock,
        mocker: MockerFixture,
        oidc_auth_flow_handler: _OIDCAuthFlowChallengeHandler,
        patch_get_auth_environment: Mock,
        received_oidc_details: Dict[str, _OIDCClientID],
        server_start_exception_cls: Type[Exception],
        states: List[str],
    ) -> None:
        """Tests that server start exceptions are reraised when awaited."""
        # Mock server starting to make it throw an exception
        mocker.patch.object(
            oidc_auth_flow_handler,
            "_start",
            autospec=True,
            side_effect=server_start_exception_cls("specific exception"),
        )
        oidc_auth_flow_handler.start_server()
        assert oidc_auth_flow_handler._server_start_task

        # Mock out receiving of OIDC details
        mock_modeller_mailbox.get_oidc_client_ids.return_value = received_oidc_details

        # Mock out code_verifier generation
        mock_secrets = mocker.patch(
            "bitfount.federated.transport.identity_verification.oidc.secrets",
            autospec=True,
        )
        mock_secrets.token_urlsafe.side_effect = code_verifiers

        # Mock out state generation
        mock_secrets.token_hex.side_effect = states

        # Mock out runner and web endpoint
        mock_runner = mocker.patch.object(
            oidc_auth_flow_handler, "_runner", autospec=True
        )
        mocker.patch.object(oidc_auth_flow_handler, "_oidc_endpoint", autospec=True)

        # Check expected server exception is re-raised
        with pytest.raises(server_start_exception_cls, match="specific exception"):
            await oidc_auth_flow_handler.handle(mock_modeller_mailbox)

        # Check server is still shutdown
        mock_runner.cleanup.assert_awaited()


@unit_test
class TestOIDCDeviceCodeHandler:
    """Tests for _OIDCDeviceCodeHandler."""

    @fixture
    def oidc_device_code_handler(
        self, audience: str, auth_domain: str, scopes: str
    ) -> _OIDCDeviceCodeHandler:
        """OIDC Device Code Flow challenge handler."""
        return _OIDCDeviceCodeHandler(auth_domain, scopes, audience)

    @fixture
    def device_code_endpoint(self, auth_domain: str) -> str:
        """Expected device code endpoint URL."""
        return f"https://{auth_domain}/oauth/device/code"

    @fixture
    def device_code_request(
        self,
        audience: str,
        client_id: str,
        scopes: str,
    ) -> _DeviceCodeRequestDict:
        """Expected device code request data."""
        return {
            "audience": audience,
            "client_id": client_id,
            "scope": scopes,
        }

    @fixture
    def device_code(self) -> str:
        """Device code."""
        return "someDeviceCode"

    @fixture
    def user_code(self) -> str:
        """User code."""
        return "someUserCode"

    @fixture
    def verification_uri(self, auth_domain: str) -> str:
        """Base verification URI."""
        return f"https://{auth_domain}/device"

    @fixture
    def verification_uri_complete(self, user_code: str, verification_uri: str) -> str:
        """Full verification URI with user code appended."""
        return f"{verification_uri}?user_code={user_code}"

    @fixture
    def expires_in(self) -> int:
        """When the device code expires."""
        return 900

    @fixture
    def interval(self) -> int:
        """The polling interval to use with the device code."""
        return 5

    @fixture
    def device_code_response(
        self,
        device_code: str,
        expires_in: int,
        interval: int,
        user_code: str,
        verification_uri: str,
        verification_uri_complete: str,
    ) -> _DeviceCodeResponseJSON:
        """Expected device code request response."""
        return {
            "device_code": device_code,
            "user_code": user_code,
            "verification_uri": verification_uri,
            "verification_uri_complete": verification_uri_complete,
            "expires_in": expires_in,
            "interval": interval,
        }

    @fixture
    def fake_now(self) -> datetime:
        """Static datetime.now() value for using in tests."""
        return datetime.now(timezone.utc)

    @fixture
    def mock_datetime_now(self, fake_now: datetime, mocker: MockerFixture) -> Mock:
        """Mock out datetime in oidc.py and return static value for now()."""
        mock_datetime = mocker.patch(
            "bitfount.federated.transport.identity_verification.oidc.datetime",
            autospec=True,
        )
        mock_datetime.now.return_value = fake_now
        return mock_datetime

    @fixture
    def expires_at(self, expires_in: int, fake_now: datetime) -> datetime:
        """The code expiration time as a datetime."""
        return fake_now + timedelta(seconds=expires_in)

    @fixture
    def expected_pod_device_code_details(
        self, device_code: str, expires_at: datetime, interval: int
    ) -> _PodDeviceCodeDetails:
        """Expected _PodDeviceCodeDetails to be returned."""
        return _PodDeviceCodeDetails(
            device_code=device_code,
            expires_at=expires_at,
            interval=interval,
        )

    @fixture
    def expected_modeller_device_code_details(
        self, user_code: str, verification_uri: str, verification_uri_complete: str
    ) -> _ModellerDeviceCodeDetails:
        """Expected _ModellerDeviceCodeDetails to be returned."""
        return _ModellerDeviceCodeDetails(
            user_code=user_code,
            verification_uri=verification_uri,
            verification_uri_complete=verification_uri_complete,
        )

    @fixture
    def expected_device_code_details_pair(
        self,
        expected_modeller_device_code_details: _ModellerDeviceCodeDetails,
        expected_pod_device_code_details: _PodDeviceCodeDetails,
    ) -> _DeviceCodeDetailsPair:
        """The pair of expected device code details."""
        return _DeviceCodeDetailsPair(
            expected_pod_device_code_details, expected_modeller_device_code_details
        )

    def test__verify_client_ids_successful(
        self,
        client_id: str,
        oidc_device_code_handler: _OIDCDeviceCodeHandler,
        patch_get_auth_environment: Mock,
    ) -> None:
        """Test _verify_client_ids() passes correctly."""
        oidc_details = {f"user{i}/pod{i}": _OIDCClientID(client_id) for i in range(3)}

        # Should just pass happily
        oidc_device_code_handler._verify_client_ids(oidc_details)

    def test__verify_client_ids_fails_diff_auth_domain(
        self,
        auth_domain: str,
        oidc_device_code_handler: _OIDCDeviceCodeHandler,
        patch_get_auth_environment: Mock,
    ) -> None:
        """Test _verify_client_ids() fails if auth domains differ."""
        # Change _auth_domain on the handler to induce mismatch
        oidc_device_code_handler._auth_domain = "not_the_same"

        with pytest.raises(
            ValueError,
            match=re.escape(
                f"Mismatch between setup auth environment and current: "
                f"expected {oidc_device_code_handler._auth_domain}, got {auth_domain}"
            ),
        ):
            oidc_device_code_handler._verify_client_ids({})

    def test__verify_client_ids_fails_diff_client_ids(
        self,
        client_id: str,
        oidc_device_code_handler: _OIDCDeviceCodeHandler,
        patch_get_auth_environment: Mock,
    ) -> None:
        """Test _verify_client_ids() fails if client IDs differ."""
        # Have one "correct" client_id
        oidc_details = {
            "user0/pod0": _OIDCClientID(client_id),
        }
        # Add two "incorrect" client_ids
        incorrect_oidc_details = {
            f"user{i}/pod{i}": _OIDCClientID(f"diff_client_id_{i}") for i in range(1, 3)
        }
        oidc_details.update(incorrect_oidc_details)

        # Create expected error message
        error_pods = "; ".join(
            [
                f"Pod ID = {pod_id}, Client ID = {details.client_id}"
                for pod_id, details in incorrect_oidc_details.items()
            ]
        )
        expected_error_message = (
            f"Authorisation environments do not match. Expected client_id "
            f'"{client_id}" but the following pods mismatched: {error_pods}'
        )

        with pytest.raises(
            AuthEnvironmentError, match=re.escape(expected_error_message)
        ):
            oidc_device_code_handler._verify_client_ids(oidc_details)

    @responses.activate
    def test__get_device_code(
        self,
        client_id: str,
        device_code_endpoint: str,
        device_code_request: _DeviceCodeRequestDict,
        device_code_response: _DeviceCodeResponseJSON,
        expected_device_code_details_pair: _DeviceCodeDetailsPair,
        mock_datetime_now: Mock,
        oidc_device_code_handler: _OIDCDeviceCodeHandler,
    ) -> None:
        """Test _get_device_code() works correctly."""
        responses.add(
            responses.POST,
            url=device_code_endpoint,
            json=device_code_response,
            match=[
                matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], device_code_request)
                )
            ],
        )

        device_code_details_pair = oidc_device_code_handler._get_device_code(client_id)

        assert device_code_details_pair == expected_device_code_details_pair

    @pytest.mark.parametrize(
        argnames=("status_code", "expected_error_msg"),
        argvalues=(
            pytest.param(
                100, "Unexpected response: status_code = 100; ", id="status_code=100"
            ),
            pytest.param(
                201, "Unexpected response: status_code = 201; ", id="status_code=201"
            ),
            pytest.param(
                300, "Unexpected response: status_code = 300; ", id="status_code=300"
            ),
            pytest.param(400, "400 Client Error: ", id="status_code=400"),
            pytest.param(500, "500 Server Error: ", id="status_code=500"),
        ),
    )
    @responses.activate
    def test__get_device_code_fails_non_200_response(
        self,
        client_id: str,
        device_code_endpoint: str,
        device_code_request: _DeviceCodeRequestDict,
        expected_error_msg: str,
        oidc_device_code_handler: _OIDCDeviceCodeHandler,
        remove_web_retry_backoff_sleep: Optional[Mock],
        status_code: int,
    ) -> None:
        """Test _get_device_code() fails on non-200 response."""
        # Check retry backoff is patched
        assert remove_web_retry_backoff_sleep is not None

        # Add a randomly filled or empty body to test different responses
        body = "not json error message" if status_code % 100 else ""
        responses.add(
            responses.POST,
            url=device_code_endpoint,
            body=body,
            status=status_code,
            match=[
                matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], device_code_request)
                )
            ],
        )

        # Modify expected message if needed
        # (i.e. relevant status code and body present)
        if status_code < 400 and body:
            expected_error_msg += body

        with pytest.raises(
            HTTPError,
            match=re.escape(expected_error_msg),
        ):
            oidc_device_code_handler._get_device_code(client_id)

        # Check retries occurred if expected
        if status_code in web_utils._RETRY_STATUS_CODES:
            assert (
                remove_web_retry_backoff_sleep.call_count
                == web_utils._DEFAULT_MAX_RETRIES
            )
        else:
            remove_web_retry_backoff_sleep.assert_not_called()

    @pytest.mark.parametrize(
        argnames=("content_type", "content_value", "expected_error"),
        argvalues=(
            pytest.param("json", "wrong json type", TypeError, id="wrong_json_type"),
            pytest.param("json", {}, KeyError, id="missing_json_key"),
            pytest.param(
                "body", "not json error message", InvalidJSONError, id="not_json"
            ),
        ),
    )
    @responses.activate
    def test__get_device_code_fails_json_errors(
        self,
        client_id: str,
        content_type: Literal["json", "body"],
        content_value: str,
        device_code_endpoint: str,
        device_code_request: _DeviceCodeRequestDict,
        expected_error: Type[Exception],
        oidc_device_code_handler: _OIDCDeviceCodeHandler,
    ) -> None:
        """Test _get_device_code() fails on JSON issues in response.

        Tests with:
            - non-dict JSON
            - missing key dict JSON
            - non-JSON response
        """
        partial_response_add = functools.partial(
            responses.add,
            responses.POST,
            url=device_code_endpoint,
            match=[
                matchers.urlencoded_params_matcher(
                    cast(Dict[str, str], device_code_request)
                )
            ],
        )
        if content_type == "json":
            partial_response_add(json=content_value)
        else:
            partial_response_add(body=content_value)

        with pytest.raises(expected_error):
            oidc_device_code_handler._get_device_code(client_id)

    @pytest.mark.parametrize("num_pods_to_use", (1, 2, 3))
    async def test_handle(
        self,
        capsys: CaptureFixture,
        device_code: str,
        expires_at: datetime,
        interval: int,
        mock_modeller_mailbox: Mock,
        mocker: MockerFixture,
        num_pods_to_use: int,
        oidc_device_code_handler: _OIDCDeviceCodeHandler,
        pod_identifiers: List[str],
        received_oidc_details: Dict[str, _OIDCClientID],
        user_code: str,
        verification_uri: str,
        verification_uri_complete: str,
    ) -> None:
        """Tests handle() method for differing numbers of pods."""
        # Patch out client ID reception
        limited_oidc_details = {
            pod_id: received_oidc_details[pod_id]
            for pod_id in pod_identifiers[:num_pods_to_use]
        }
        mock_modeller_mailbox.get_oidc_client_ids.return_value = limited_oidc_details

        # Patch out verification check
        mock_verify_client_ids = mocker.patch.object(
            oidc_device_code_handler, "_verify_client_ids", autospec=True
        )

        # Patch out device code retrieval
        device_code_details = [
            _DeviceCodeDetailsPair(
                _PodDeviceCodeDetails(
                    device_code=f"{device_code}_{i}",
                    expires_at=expires_at,
                    interval=interval,
                ),
                _ModellerDeviceCodeDetails(
                    user_code=f"{user_code}_{i}",
                    verification_uri=verification_uri,
                    verification_uri_complete=f"{verification_uri_complete}_{i}",
                ),
            )
            for i in range(num_pods_to_use)
        ]
        mocker.patch.object(
            oidc_device_code_handler,
            "_get_device_code",
            side_effect=device_code_details,
        )

        # Mock out webbrowser
        mock_webbrowser = mocker.patch(
            "bitfount.federated.transport.identity_verification.oidc.webbrowser",
            autospec=True,
        )

        # Mock out asyncio.sleep()
        mock_asyncio_sleep = mocker.patch.object(asyncio, "sleep", autospec=True)

        # Perform call
        await oidc_device_code_handler.handle(mock_modeller_mailbox)

        # Check verification occurred
        mock_verify_client_ids.assert_called_once_with(limited_oidc_details)

        # Check pod details sent as expected
        mock_modeller_mailbox.send_oidc_device_code_responses.assert_awaited_once_with(
            {
                pod_identifiers[i]: device_code_details[i].pod_details
                for i in range(num_pods_to_use)
            }
        )

        # Check URLs printed
        stdout = capsys.readouterr().out
        if num_pods_to_use == 1:
            pod_id = pod_identifiers[0]
            modeller_details = device_code_details[0].modeller_details
            user_code = modeller_details.user_code
            verification_uri_complete = modeller_details.verification_uri_complete
            assert (
                "A browser window will be opened, please login and confirm "
                "identity verification access for the pod.\n"
                "If a browser window is not opened, then please visit the "
                "following URL:\n"
                f"\tFor {pod_id}: code: {user_code}; {verification_uri_complete}"
            ) in stdout
        else:
            url_details = []
            for i in range(num_pods_to_use):
                pod_id = pod_identifiers[i]
                modeller_details = device_code_details[i].modeller_details
                user_code = modeller_details.user_code
                verification_uri_complete = modeller_details.verification_uri_complete
                url_details.append(
                    f"\tFor {pod_id}: code: {user_code}; {verification_uri_complete}"
                )
            assert (
                "Browser windows will be opened, please login and confirm "
                "identity verification access for these pods.\n"
                "If no or not all browser windows open, then please visit "
                "the following URLs:\n" + "\n".join(url_details)
            ) in stdout

        # Check sleep was called
        mock_asyncio_sleep.assert_awaited_once_with(1)

        # Check URLs opened correctly
        assert mock_webbrowser.open_new_tab.call_args_list == [
            call(device_code_details[i].modeller_details.verification_uri_complete)
            for i in range(num_pods_to_use)
        ]



"""Custom exceptions for the data package."""
from bitfount.exceptions import BitfountError


class BitfountSchemaError(BitfountError):
    """Errors related to BitfountSchema."""

    pass


class DataStructureError(BitfountError):
    """Errors related to Datastructure."""

    pass


class DataSourceError(BitfountError):
    """Errors related to Datasource."""

    pass


class DatasetSplitterError(BitfountError):
    """Errors related to DatasetSplitter."""

    pass


class DatabaseSchemaNotFoundError(BitfountError):
    """Raised when a specified database schema is not found."""

    pass


class DatabaseMissingTableError(BitfountError):
    """Raised when a specified database table is not found."""

    pass


class DatabaseUnsupportedQueryError(BitfountError):
    """Raised when an unsupported database query is provided."""

    pass


class DataNotLoadedError(BitfountError):
    """Raised if a data operation is attempted prior to data loading.

    This is usually raised because `load_data` has not been called yet.
    """

    pass


class DuplicateColumnError(BitfountError):
    """Raised if the column names are duplicated in the data.

    This can be raised by the sql algorithms with multi-table pods.
    """

    pass


class ExcelSourceError(BitfountError):
    """Error for ExcelSources.

    We raise this if trying to get the types of the
    columns with no table name in a multi-table datasource.
    """

    pass



"""Test protocol.py with pytorch backend."""
import copy
from typing import List, Optional, Union, cast
from unittest.mock import AsyncMock, Mock, create_autospec

from _pytest.logging import LogCaptureFixture
import pytest
from pytest import fixture
from pytest_mock import MockerFixture

from bitfount.backends.pytorch.federated.shim import PyTorchBackendTensorShim
from bitfount.backends.pytorch.models.models import PyTorchTabularClassifier
from bitfount.data.datasources.dataframe_source import DataFrameSource
from bitfount.data.datastructure import DataStructure
from bitfount.data.schema import BitfountSchema
from bitfount.federated.aggregators.aggregator import (
    Aggregator,
    _ModellerSide,
    _WorkerSide,
)
from bitfount.federated.aggregators.secure import SecureAggregator
from bitfount.federated.algorithms.model_algorithms import federated_training
from bitfount.federated.algorithms.model_algorithms.federated_training import (
    FederatedModelTraining,
    _BaseModelTrainingMixIn,
)
from bitfount.federated.protocols.model_protocols.federated_averaging import (
    FederatedAveraging,
    _FederatedAveragingCompatibleAlgoFactory,
)
from bitfount.federated.secure import SecureShare
from bitfount.federated.transport.modeller_transport import _ModellerMailbox
from bitfount.federated.transport.worker_transport import _WorkerMailbox
from bitfount.types import DistributedModelProtocol, _SerializedWeights, _StrAnyDict
from tests.bitfount import TEST_SECURITY_FILES
from tests.utils import PytestRequest
from tests.utils.helper import (
    backend_test,
    create_dataset,
    create_datastructure,
    create_schema,
    unit_test,
)


@fixture
def datasource() -> DataFrameSource:
    """Returns datasource."""
    dataset = create_dataset()
    return DataFrameSource(dataset)


@fixture
def datastructure_mult_pods() -> DataStructure:
    """Fixture for datastructure."""
    ds = create_datastructure()
    ds.table = {
        "bitfount/census-income": "census-income",
        "bitfount/census-income-2": "census-income-2",
    }
    return ds


@fixture
def datastructure() -> DataStructure:
    """Fixture for datastructure."""
    return create_datastructure()


@fixture
def schema() -> BitfountSchema:
    """Fixture for schema."""
    return create_schema(classification=True)


@fixture
def aggregator(request: PytestRequest) -> Union[Aggregator, SecureAggregator]:
    """Returns mock aggregator."""
    # request.param denotes whether we should use secure aggregator
    if request.param:
        return SecureAggregator(
            secure_share=SecureShare(),
        )
    return Aggregator()


@fixture
def mock_aggregator() -> Mock:
    """A mock that implements the aggregator (factory) interface."""
    mock_aggregator: Mock = create_autospec(spec=Aggregator, instance=True)

    mock_modeller_side = create_autospec(spec=_ModellerSide, instance=True)
    mock_aggregator.modeller.return_value = mock_modeller_side

    mock_worker_side = create_autospec(spec=_WorkerSide, instance=True)
    mock_aggregator.worker.return_value = mock_worker_side
    # Autospec is not auto setting this as async for some reason, so we do it manually.
    mock_worker_side.run = AsyncMock()

    return mock_aggregator


@fixture
def mock_modeller_mailbox() -> Mock:
    """Returns mock mailbox."""
    mock_mailbox: Mock = create_autospec(_ModellerMailbox, instance=True)
    return mock_mailbox


@fixture
def model(datastructure_mult_pods: DataStructure) -> PyTorchTabularClassifier:
    """Returns distributed model."""
    return PyTorchTabularClassifier(
        datastructure=datastructure_mult_pods, schema=BitfountSchema(), epochs=1
    )


@fixture
def federated_algorithm_with_model(
    model: PyTorchTabularClassifier,
) -> FederatedModelTraining:
    """Returns federated algorithm with embedded distributed model."""
    return FederatedModelTraining(model=model)


@fixture
def federated_algorithm(
    datastructure: DataStructure, request: PytestRequest
) -> FederatedModelTraining:
    """Returns federated algorithm with configurable iterations."""
    epochs, steps = request.param
    schema = create_schema(classification=True)
    return FederatedModelTraining(
        model=PyTorchTabularClassifier(
            datastructure=datastructure,
            schema=schema,
            epochs=epochs,
            steps=steps,
        )
    )


@fixture
def mock_federated_algorithm(request: PytestRequest) -> Mock:
    """Creates a mock that matches the federated algorithm.

    Uses request parameters to set the number of steps/epochs.
    """
    epochs, steps = request.param

    mock_algo_factory: Mock = create_autospec(
        _FederatedAveragingCompatibleAlgoFactory, instance=True
    )

    mock_worker_side: Mock = create_autospec(
        federated_training._WorkerSide, instance=True
    )
    mock_algo_factory.worker.return_value = mock_worker_side

    mock_modeller_side = create_autospec(
        federated_training._ModellerSide, instance=True
    )
    mock_algo_factory.modeller.return_value = mock_modeller_side
    mock_modeller_side.epochs = epochs
    mock_modeller_side.steps = steps

    return mock_algo_factory


@fixture
def mock_modeller_run_method(mocker: MockerFixture) -> AsyncMock:
    """Mocks out the Modeller.run() method in protocol.py."""
    mock_run_method: AsyncMock = mocker.patch(
        "bitfount.federated.protocols.base._Modeller.run"
    )
    return mock_run_method


@fixture
def pod_identifiers() -> List[str]:
    """A list of pod identifiers."""
    return ["bitfount/census-income", "bitfount/census-income-2"]


@backend_test
class TestFederatedAveraging:
    """Tests FederatedAveraging protocol with pytorch models."""

    # TODO: [BIT-983] Should these tests mostly be on the main protocol tests
    #       rather than backend? Or should they be @integration_test marked?

    @unit_test
    @pytest.mark.parametrize(
        "federated_algorithm, epochs_between_parameter_updates, steps_between_parameter_updates",  # noqa: B950
        [
            # The first tuple is the (epochs, steps) args for federated_algorithm
            ((3, None), None, 1),
            ((None, 30), 10, None),
            ((None, 30), None, 31),
            ((30, None), 31, None),
        ],
        indirect=["federated_algorithm"],
    )
    @pytest.mark.parametrize("aggregator", [False], indirect=True)
    def test_iterations_incompatible_raises_value_error(
        self,
        aggregator: Union[Aggregator, SecureAggregator],
        epochs_between_parameter_updates: Optional[int],
        federated_algorithm: FederatedModelTraining,
        mock_modeller_mailbox: Mock,
        steps_between_parameter_updates: Optional[int],
    ) -> None:
        """Ensure algorithm steps/epochs is the same as model."""
        with pytest.raises(ValueError):
            protocol_factory = FederatedAveraging(
                algorithm=federated_algorithm,
                aggregator=aggregator,
                steps_between_parameter_updates=steps_between_parameter_updates,
                epochs_between_parameter_updates=epochs_between_parameter_updates,
            )
            protocol = protocol_factory.modeller(mailbox=mock_modeller_mailbox)
            protocol.perform_iterations_checks()

    @unit_test
    @pytest.mark.parametrize(
        "mock_federated_algorithm, epochs_between_parameter_updates, steps_between_parameter_updates, expected_result",  # noqa: B950
        [
            # The first tuple is the (epochs, steps) args for federated_algorithm
            ((10, None), 1, None, 10),
            ((10, None), 3, None, 3),
            ((None, 10), None, 2, 5),
            ((None, 20), None, 7, 2),
        ],
        indirect=["mock_federated_algorithm"],
    )
    def test_get_num_federated_iterations(
        self,
        epochs_between_parameter_updates: Optional[int],
        expected_result: int,
        mock_aggregator: Mock,
        mock_federated_algorithm: FederatedModelTraining,
        mock_modeller_mailbox: Mock,
        steps_between_parameter_updates: Optional[int],
    ) -> None:
        """Assert number of federated iterations is calculated correctly."""
        # TODO: [BIT-983] This should probably be a test on protocol, rather than on the
        #       backend, as it's testing core functionality.
        protocol_factory = FederatedAveraging(
            algorithm=mock_federated_algorithm,
            aggregator=mock_aggregator,
            steps_between_parameter_updates=steps_between_parameter_updates,
            epochs_between_parameter_updates=epochs_between_parameter_updates,
        )
        protocol = protocol_factory.modeller(mailbox=mock_modeller_mailbox)
        assert protocol.get_num_federated_iterations() == expected_result

    @unit_test
    @pytest.mark.parametrize(
        "federated_algorithm, epochs_between_parameter_updates, steps_between_parameter_updates",  # noqa: B950
        [
            # The first tuple is the (epochs, steps) args for federated_algorithm
            ((10, None), 1, None),
            ((10, None), 3, None),
            ((None, 10), None, 2),
            ((None, 20), None, 7),
        ],
        indirect=["federated_algorithm"],
    )
    async def test_worker_model_training_for_appropriate_amount(
        self,
        datasource: DataFrameSource,
        epochs_between_parameter_updates: Optional[int],
        federated_algorithm: FederatedModelTraining,
        mock_aggregator: Mock,
        mocker: MockerFixture,
        schema: BitfountSchema,
        steps_between_parameter_updates: Optional[int],
    ) -> None:
        """Test underlying worker model iterations updated correctly during training."""
        # Create mailbox and mock out relevant functions
        mock_worker_mailbox = create_autospec(_WorkerMailbox, instance=True)
        mock_get_model_updates = mocker.patch(
            "bitfount.federated.protocols.model_protocols.federated_averaging._get_model_parameters"
        )
        mock_send_model_updates = mocker.patch(
            "bitfount.federated.protocols.model_protocols.federated_averaging._send_parameter_update"
        )
        # Mock out so training complete never returns truthy value prematurely
        mock_worker_mailbox.get_training_iteration_complete_update.return_value = False

        # Create protocol and initialise model
        protocol_factory = FederatedAveraging(
            algorithm=federated_algorithm,
            aggregator=mock_aggregator,
            steps_between_parameter_updates=steps_between_parameter_updates,
            epochs_between_parameter_updates=epochs_between_parameter_updates,
        )
        protocol = protocol_factory.worker(mailbox=mock_worker_mailbox, hub=Mock())
        protocol.algorithm.initialise(datasource=datasource)
        num_federated_iterations = protocol.get_num_federated_iterations()
        expected_total_num_iterations = num_federated_iterations * cast(
            int, steps_between_parameter_updates or epochs_between_parameter_updates
        )

        async def mock_get_model_updates_coroutine(
            _mailbox: _WorkerMailbox,
        ) -> _SerializedWeights:
            """Mock out get_model_parameters."""
            params = cast(
                _BaseModelTrainingMixIn, protocol.algorithm
            ).get_param_states()
            model_params = copy.deepcopy(cast(_StrAnyDict, params))
            for name, param in model_params.items():
                model_params[name] = PyTorchBackendTensorShim.to_list(param)
            return model_params

        mock_get_model_updates.side_effect = mock_get_model_updates_coroutine

        await protocol.run(datasource=datasource)

        # We expect to send and receive updates in each federated iteration (and
        # receive one extra "update" at the very beginning).
        assert mock_send_model_updates.call_count == num_federated_iterations

        # One initial set of parameters and then one per federated iteration
        assert mock_get_model_updates.await_count == num_federated_iterations + 1

        # Check that the iterations are correct, whichever form we're using.
        # If we are using epochs, we can only check that the epoch number is right on
        # the final iteration. If we are using steps, instead we check that the total
        # number of iterations is correct.
        assert isinstance(federated_algorithm.model, DistributedModelProtocol)
        if epochs_between_parameter_updates:
            # The `current_epoch` attribute denotes the epoch number of the epoch
            # about to be trained on the next iteration (starting from 0)
            assert (
                federated_algorithm.model._pl_trainer.current_epoch  # type: ignore[attr-defined] # Reason: Model has _pl_trainer # noqa: B950
                == epochs_between_parameter_updates
            )
        else:
            assert (
                expected_total_num_iterations
                == federated_algorithm.model._total_num_batches_trained
            )

    @unit_test
    def test_helper_run_method_with_algorithm(
        self,
        federated_algorithm_with_model: FederatedModelTraining,
        mock_bitfount_session: Mock,
        mock_modeller_run_method: Mock,
        pod_identifiers: List[str],
    ) -> None:
        """Tests protocol helper run method with algorithm."""
        protocol = FederatedAveraging(algorithm=federated_algorithm_with_model)

        protocol.run(
            pod_identifiers=pod_identifiers,
            private_key_or_file=TEST_SECURITY_FILES / "test_private.testkey",
        )

        mock_modeller_run_method.assert_called_once_with(
            pod_identifiers, require_all_pods=False
        )

    @unit_test
    def test_helper_run_method_with_model(
        self,
        mock_bitfount_session: Mock,
        mock_modeller_run_method: Mock,
        model: PyTorchTabularClassifier,
        pod_identifiers: List[str],
    ) -> None:
        """Tests helper run method with a model."""
        protocol = FederatedAveraging(algorithm=FederatedModelTraining(model=model))
        protocol.run(
            pod_identifiers=pod_identifiers,
            private_key_or_file=TEST_SECURITY_FILES / "test_private.testkey",
        )

        mock_modeller_run_method.assert_called_once_with(
            pod_identifiers, require_all_pods=False
        )

    @unit_test
    def test_helper_run_method_with_model_and_algorithm(
        self,
        caplog: LogCaptureFixture,
        federated_algorithm_with_model: FederatedModelTraining,
        mock_bitfount_session: Mock,
        mock_modeller_run_method: Mock,
        pod_identifiers: List[str],
    ) -> None:
        """Tests helper run method with an algorithm and a model.

        This tests that the run method will still run but that it just issues a warning
        regarding the extra model argument.
        """
        mock_model = Mock()
        protocol = FederatedAveraging(
            model=mock_model, algorithm=federated_algorithm_with_model
        )
        protocol.run(
            pod_identifiers=pod_identifiers,
            private_key_or_file=TEST_SECURITY_FILES / "test_private.testkey",
        )

        mock_modeller_run_method.assert_called_once_with(
            pod_identifiers, require_all_pods=False
        )
        mock_model.assert_not_called()
        mock_model.backend_tensor_shim.assert_not_called()
        assert "Ignoring provided model. Algorithm already has a model." in caplog.text



"""Tests plugin functionality."""

from importlib import import_module, reload
import inspect
from pathlib import Path
import time

from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
from _pytest.tmpdir import TempPathFactory
from pytest import fixture
from pytest_mock import MockerFixture

import bitfount
from bitfount.runners.config_schemas import PodConfig, PodDataConfig
from bitfount.runners.pod_runner import setup_pod_from_config
from tests.utils.helper import unit_test

BITFOUNT_DATA_MODULE_PATH = "bitfount.data"
BITFOUNT_DATASOURCES_MODULE_PATH = f"{BITFOUNT_DATA_MODULE_PATH}.datasources"
DUMMY_MODULE_NAME = "dummy_module"
DUMMY_SOURCE_NAME = "DummySource"


@unit_test
class TestPlugins:
    """Unified class for testing all plugin functionality."""

    @fixture(autouse=True)
    def patch_plugins_dir(
        self, monkeypatch: MonkeyPatch, tmp_path_factory: TempPathFactory
    ) -> None:
        """Monkeypatch `BITFOUNT_PLUGIN_PATH` in bitfount.config to temporary directory.

        We must do it this way instead of by monkeypatching the `BITFOUNT_HOME`
        environment variable to avoid having to reload the `bitfount.config` module as
        this causes issues with the cache of the `_get_environment` function.
        """
        tmpdir = str(tmp_path_factory.mktemp("temp", numbered=True))
        monkeypatch.setattr(
            "bitfount.config.BITFOUNT_PLUGIN_PATH",
            Path(tmpdir) / ".bitfount/_plugins",
        )

    @fixture(scope="function")
    def plugin_suffix(self) -> str:
        """Generates a unique suffix for each plugin in this test run."""
        return str(int(time.time_ns()))  # unix nanosecond timestamp

    @fixture(scope="function")
    def dummy_module_name(self, plugin_suffix: str) -> str:
        """Generates a unique plugin module name for each plugin in this test run."""
        return f"{DUMMY_MODULE_NAME}_{plugin_suffix}"

    @fixture(scope="function")
    def dummy_source_plugin_name(self, plugin_suffix: str) -> str:
        """Generates a unique plugin source name for each plugin in this test run."""
        return f"{DUMMY_SOURCE_NAME}_{plugin_suffix}"

    @fixture(scope="function")
    def dummy_plugin(self, dummy_source_plugin_name: str) -> str:
        """Returns a dummy plugin class."""
        return inspect.cleandoc(
            f"""
        import os
        from typing import Any, Dict, Iterable, List, Optional, Union

        import numpy as np
        import pandas as pd
        from pydantic import AnyUrl

        from bitfount.data.datasources.base_source import BaseSource
        from bitfount.types import _Dtypes

        class {dummy_source_plugin_name}:

            def __init__(
                self,
                path: Union[os.PathLike, AnyUrl, str],
                read_excel_kwargs: Optional[Dict[str, Any]] = None,
                **kwargs: Any,
            ):
                ...

            def get_data(self, **kwargs: Any) -> pd.DataFrame:
                ...

            def get_values(
                self, col_names: List[str], **kwargs: Any
            ) -> Dict[str, Iterable[Any]]:
                ...

            def get_column(
                self, col_name: str, **kwargs: Any
            ) -> Union[np.ndarray, pd.Series]:
                ...

            def get_dtypes(self, **kwargs: Any) -> _Dtypes:
                ...

            def __len__(self) -> int:
                ...

            @property
            def multitable(self) -> bool: # intentional misspelling
                ...
        """
        )

    @fixture(scope="function")
    def dummy_abstract_source_plugin(
        self, dummy_plugin: str, dummy_source_plugin_name: str
    ) -> str:
        """Returns a class that subclasses but doesn't fully implement BaseSource."""
        return dummy_plugin.replace(
            dummy_source_plugin_name, f"{dummy_source_plugin_name}(BaseSource)"
        )

    @fixture(scope="function")
    def dummy_source_plugin(self, dummy_abstract_source_plugin: str) -> str:
        """Returns a class that fully implements BaseSource."""
        return dummy_abstract_source_plugin.replace("multitable", "multi_table")

    def test_no_plugins_dir_created(self) -> None:
        """Tests that the plugin directory is created if it doesn't exist.

        Also tests that it's okay for there to be no plugins within the directory.
        """
        assert bitfount.config.BITFOUNT_PLUGIN_PATH.exists() is False
        reload(bitfount.data.datasources)
        assert bitfount.config.BITFOUNT_PLUGIN_PATH.exists() is True

    def test_plugin_in_wrong_dir_gets_ignored(
        self, dummy_module_name: str, dummy_source_plugin: str
    ) -> None:
        """Tests that plugins in the wrong directory are ignored."""
        reload(bitfount.data.datasources)  # Creates the plugin directory
        # `data_sources` is intentionally the wrong directory, should be `datasources`
        dummy_plugin_path = bitfount.config.BITFOUNT_PLUGIN_PATH / "data_sources"
        dummy_plugin_path.mkdir()
        (dummy_plugin_path / f"{dummy_module_name}.py").touch()
        (dummy_plugin_path / f"{dummy_module_name}.py").write_text(dummy_source_plugin)
        reload(bitfount.data)
        reload(bitfount.data.datasources)
        datasources = import_module("bitfount.data.datasources")
        assert not hasattr(datasources, dummy_module_name)

    def test_datasource_plugin_that_doesnt_subclass_base_source_gets_ignored(
        self, dummy_module_name: str, dummy_plugin: str, dummy_source_plugin_name: str
    ) -> None:
        """Tests that plugins that don't subclass BaseSource are ignored."""
        reload(bitfount.data.datasources)  # Creates the plugin directory
        dummy_plugin_path = bitfount.config.BITFOUNT_PLUGIN_PATH / "datasources"
        (dummy_plugin_path / f"{dummy_module_name}.py").touch()
        (dummy_plugin_path / f"{dummy_module_name}.py").write_text(dummy_plugin)
        reload(bitfount.data)
        reload(bitfount.data.datasources)

        # Tests that the module is still imported but the class is not
        datasources = import_module(BITFOUNT_DATASOURCES_MODULE_PATH)
        data = import_module(BITFOUNT_DATA_MODULE_PATH)
        assert hasattr(datasources, dummy_module_name)
        assert not hasattr(data, dummy_source_plugin_name)

    def test_datasource_plugin_that_doesnt_implement_base_source_gets_ignored(
        self,
        caplog: LogCaptureFixture,
        dummy_abstract_source_plugin: str,
        dummy_module_name: str,
        dummy_source_plugin_name: str,
    ) -> None:
        """Tests that plugins that don't implement BaseSource are ignored."""
        caplog.set_level("DEBUG")
        reload(bitfount.data.datasources)  # Creates the plugin directory
        dummy_plugin_path = bitfount.config.BITFOUNT_PLUGIN_PATH / "datasources"
        (dummy_plugin_path / f"{dummy_module_name}.py").touch()
        (dummy_plugin_path / f"{dummy_module_name}.py").write_text(
            dummy_abstract_source_plugin
        )
        reload(bitfount.data)
        reload(bitfount.data.datasources)

        # Tests that the module is still imported but the class is not
        datasources = import_module(BITFOUNT_DATASOURCES_MODULE_PATH)
        data = import_module(BITFOUNT_DATA_MODULE_PATH)
        assert hasattr(datasources, dummy_module_name)
        assert not hasattr(data, dummy_source_plugin_name)

        # Tests that the appropriate debug message is logged
        assert caplog.records[0].levelname == "DEBUG"
        assert caplog.records[0].message == (
            f"Found class {dummy_source_plugin_name} in module {dummy_module_name} which did not fully implement BaseSource. "  # noqa: B950
            "Skipping."
        )

    def test_datasource_plugin_gets_loaded_via_api(
        self,
        dummy_module_name: str,
        dummy_source_plugin: str,
        dummy_source_plugin_name: str,
    ) -> None:
        """Tests that plugins that implement BaseSource are loaded via the API."""
        reload(bitfount.data.datasources)  # Creates the plugin directory
        dummy_plugin_path = bitfount.config.BITFOUNT_PLUGIN_PATH / "datasources"
        (dummy_plugin_path / f"{dummy_module_name}.py").touch()
        (dummy_plugin_path / f"{dummy_module_name}.py").write_text(dummy_source_plugin)
        reload(bitfount.data.datasources)
        reload(bitfount.data)

        # Tests that the module and class can both be imported
        datasources = import_module(BITFOUNT_DATASOURCES_MODULE_PATH)
        data = import_module(BITFOUNT_DATA_MODULE_PATH)
        assert hasattr(datasources, dummy_module_name)
        assert hasattr(data, dummy_source_plugin_name)

    def test_datasource_plugin_gets_loaded_via_yaml(
        self,
        dummy_module_name: str,
        dummy_source_plugin: str,
        dummy_source_plugin_name: str,
        mocker: MockerFixture,
    ) -> None:
        """Tests that plugins that implement BaseSource are loaded via YAML."""
        reload(bitfount.data.datasources)  # Creates the plugin directory
        dummy_plugin_path = bitfount.config.BITFOUNT_PLUGIN_PATH / "datasources"
        (dummy_plugin_path / f"{dummy_module_name}.py").touch()
        (dummy_plugin_path / f"{dummy_module_name}.py").write_text(dummy_source_plugin)
        reload(bitfount.data)
        reload(bitfount.data.datasources)

        mocker.patch("bitfount.runners.pod_runner._create_bitfounthub")
        mocker.patch("bitfount.runners.pod_runner._create_access_manager")
        mocker.patch("bitfount.runners.pod_runner._get_pod_keys")
        mock_pod = mocker.patch("bitfount.runners.pod_runner.Pod")

        # Sets up the YAML config
        pod_config = PodConfig(
            datasource=dummy_source_plugin_name,
            name="dummy_pod",
            data_config=PodDataConfig(datasource_args={"path": "dummy_module.csv"}),
        )

        # Creates the pod
        setup_pod_from_config(pod_config)

        # Checks that the pod was created
        mock_pod.assert_called_once()

        # Checks that the pod was created with the correct datasource argument
        assert (
            mock_pod.call_args.kwargs["datasource"].__class__.__name__
            == dummy_source_plugin_name
        )



"""Parsing functionaity for Transformation instances.

This module contains the transformations parser that will parse a set of
Transformation specifications from a YAML file and create a set of
Transformation instances.
"""
from collections import defaultdict
from os import PathLike
from typing import DefaultDict, Dict, Iterable, List, Mapping, Set, Tuple, Union, cast

import yaml

from bitfount.transformations.base_transformation import (
    TRANSFORMATION_REGISTRY,
    MultiColumnOutputTransformation,
    Transformation,
)
from bitfount.transformations.exceptions import TransformationParsingError
from bitfount.transformations.references import (
    _COLUMN_REFERENCE,
    _TRANSFORMATION_REFERENCE,
)


class TransformationsParser:
    """A parser for converting YAML configs into `Transformation` instances.

    :::info

    To refer to a column, we can use `c`, `col` or `column` as a prefix followed by the
    column name.

    To refer to a transformation, we can use `t`, `tran` or `transformation` as a prefix
    followed by the column name.

    :::
    """

    def parse(self, yaml_str: str) -> Tuple[List[Transformation], Set[str]]:
        """Parses a transformations YAML config into a list of Transformation objects.

        Args:
            yaml_str: The YAML config as a string.

        Returns: A tuple of the list of Transformation instances and any column names
            that are referenced.

        Raises:
            TransformationParserError: If parsing errors occur. The errors are stored on
                the `errors` attribute of the exception.
        """
        loaded = yaml.safe_load(yaml_str)
        return self.deserialize_transformations(loaded["Transformations"])

    def parse_file(
        self, path: Union[str, PathLike]
    ) -> Tuple[List[Transformation], Set[str]]:
        """Parses a transformations YAML file into a list of Transformation objects.

        Args:
            path: Path to the YAML config file.

        Returns: A tuple of the list of Transformation instances and any column names
            that are referenced.

        Raises:
            TransformationParserError: If parsing errors occur. The errors are stored on
                the `errors` attribute of the exception.
        """
        with open(path) as f:
            return self.parse(f.read())

    def deserialize_transformations(
        self, data: Iterable[Mapping[str, Mapping]]
    ) -> Tuple[List[Transformation], Set[str]]:
        """Deserializes a list of python objects into `Transformation` instances.

        Args:
            data: The serialized data representing the list of Transformations.

        Raises:
            TransformationParserError: If parsing errors occur. The errors are stored
                on the `errors` attribute of the exception.

        Returns:
            A tuple of the list of deserialized `Transformation` instances and any
            column names that are referenced.
        """
        transformations = []
        errors = []

        # Data should be a list/iterable of mappings of
        # transformation name -> transformation args
        for i, t_spec in enumerate(data):
            try:
                transformation = self._deserialize_transformation(i, t_spec)
                transformations.append(transformation)
            except TransformationParsingError as tpe:
                errors.extend(tpe.errors)

        # Check that each transformation has a unique name
        try:
            self._check_names_unique(transformations)
        except TransformationParsingError as tpe:
            errors.extend(tpe.errors)

        # Hook transformation references together, ensuring they are
        # correctly referenced
        try:
            self._hook_transformations_together(transformations)
        except TransformationParsingError as tpe:
            errors.extend(tpe.errors)

        # Extract the set of column names that are referenced
        col_refs = self._extract_column_refs(transformations)

        if errors:
            raise TransformationParsingError(errors)

        return transformations, col_refs

    def _deserialize_transformation(
        self, t_spec_idx: int, t_spec: Union[str, Mapping[str, Mapping]]
    ) -> Transformation:
        """Deserializes an individual transformation.

        Args:
            t_spec_idx:
                The index of the transformation spec in the list.
            t_spec:
                The mapping representing the transformation information.

        Raises:
            TransformationParserError:
                If the spec contains more than one mapping or if no transformation
                of that type is registered.

        Returns:
            The deserialized transformation.
        """
        # If t_spec isn't a key, convert it into one with no values
        if not isinstance(t_spec, Mapping):
            t_spec = {t_spec: {}}

        # Check that each mapping is only for a single transformation
        if len(t_spec) != 1:
            raise TransformationParsingError(
                f"Each transformation mapping must contain exactly one "
                f"transformation; mapping {t_spec_idx} contains {len(t_spec)}."
            )

        # Retrieve transformation schema from registry and deserialize
        t_name, t_details = list(t_spec.items())[0]  # as only one long
        t_name = t_name.lower()
        try:
            schema = TRANSFORMATION_REGISTRY[t_name].schema()
        except KeyError:
            raise TransformationParsingError(
                f"No transformation registered with name {t_name}."
            )
        transformation: Transformation = schema.load(t_details)
        return transformation

    def _check_names_unique(self, transformations: Iterable[Transformation]) -> None:
        """Checks that each transformation has a unique name.

        Args:
            transformations:
                List of transformation instances.

        Raises:
            TransformationParsingError:
                If duplicate names are detected.
        """
        errors = []
        seen = set()
        duplicated = set()
        multi_column_outputs = []

        for t in transformations:
            # Save multi-column output transformations to deal with their
            # additional columns later
            if isinstance(t, MultiColumnOutputTransformation):
                multi_column_outputs.append(t)

            # Check exact name conflicts
            name = t.name
            if name not in seen:
                seen.add(name)
            else:
                duplicated.add(name)
        if duplicated:
            errors.extend(
                [f"Duplicate transformation name: {dupe}." for dupe in duplicated]
            )

        # Check multi-column output transformations for clashes
        mco_clashes: DefaultDict[str, List[str]] = defaultdict(list)
        for t in multi_column_outputs:
            name = cast(str, t.name)
            cols = set(t.columns)

            # Find clashes against `seen` set.
            clashes = cols.intersection(seen)
            if clashes:
                mco_clashes[name].extend(clashes)
            # Regardless, need to update `seen` with the new cols
            seen.update(cols)
        if mco_clashes:
            errors.extend(
                [
                    f"Multi-column output clash: {clash} "
                    f"(from output column of {name})."
                    for name, clashes in mco_clashes.items()
                    for clash in clashes
                ]
            )

        # Raise any errors that occur
        if errors:
            raise TransformationParsingError(errors)

    def _hook_transformations_together(
        self, transformations: Iterable[Transformation]
    ) -> None:
        """Replaces transformation name references with the actual transformations.

        Iterates through all Transformations, replacing any arguments that reference
        other Transformations by name with the actual Transformations.

        Args:
            transformations:
                The list of transformations.

        Raises:
            TransformationParsingError:
                If references are made to transformations that have yet to be defined
                in the order.
        """
        errors = []
        transforms_map: Dict[str, Transformation] = dict()

        # In order, check that any references to previous transformations are
        # correct and exist.
        for t in transformations:
            t.name = cast(str, t.name)
            transforms_map[t.name] = t

            # Check if any attributes reference transformations
            for attr, value in vars(t).items():
                try:
                    match = _TRANSFORMATION_REFERENCE.fullmatch(value)
                    # If referencing a transformation, replace the attribute with
                    # the actual transformation instance. Raise exception if the
                    # transformation hasn't been seen yet.
                    if match:
                        ref = match.group(1)
                        try:
                            setattr(t, attr, transforms_map[ref])
                        except KeyError:
                            errors.append(
                                f"Transformation, {t.name}, attempted to use "
                                f'transformation "{ref}" before it was defined.'
                            )
                except TypeError:
                    pass

        if errors:
            raise TransformationParsingError(errors)

    def _extract_column_refs(
        self, transformations: Iterable[Transformation]
    ) -> Set[str]:
        """Extracts the names of any columns referenced in the transformations.

        Args:
            transformations:
                The list of transformations.

        Returns:
            The set of column names referenced in transformation attributes.
        """
        col_refs = set()

        def add_col_ref(string: str) -> None:
            match = _COLUMN_REFERENCE.fullmatch(string)
            if match:
                col_refs.add(match.group(1))

        for t in transformations:
            for _attr, value in vars(t).items():
                try:
                    if isinstance(value, list):
                        for value_ in value:
                            add_col_ref(value_)
                    else:
                        add_col_ref(value)
                except TypeError:
                    pass
        return col_refs
